"""Model for DepositPreauth transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

from typing_extensions import Self

from xrpl.models.nested_model import NestedModel
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import (
    KW_ONLY_DATACLASS,
    MAX_CREDENTIAL_ARRAY_LENGTH,
    require_kwargs_on_init,
)


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class DepositPreauth(Transaction):
    """
    Represents a `DepositPreauth <https://xrpl.org/depositpreauth.html>`_
    transaction, which gives another account pre-approval to deliver payments to
    the sender of this transaction, if this account is using
    `Deposit Authorization <https://xrpl.org/depositauth.html>`_.
    """

    authorize: Optional[str] = None
    """
    Grant preauthorization to this address. You must provide this OR
    ``unauthorize`` but not both.
    """

    unauthorize: Optional[str] = None
    """
    Revoke preauthorization from this address. You must provide this OR
    ``authorize`` but not both.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.DEPOSIT_PREAUTH,
        init=False,
    )

    authorize_credentials: Optional[List[Credential]] = None
    """The credential(s) that received the preauthorization. (Any account with these
    credentials can send preauthorized payments)."""

    unauthorize_credentials: Optional[List[Credential]] = None
    """The credential(s) whose preauthorization should be revoked."""

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if self.authorize and self.unauthorize:
            errors["DepositPreauth"] = (
                "One of authorize and unauthorize must be set, not both."
            )

        if (
            self.authorize is None
            and self.unauthorize is None
            and self.authorize_credentials is None
            and self.unauthorize_credentials is None
        ):
            errors["DepositPreauth"] = (
                "Exactly one field must be set for either authorize, unauthorize, "
                + "authorize_credentials or unauthorize_credentials. It is "
                + "invalid if none of the params are specified."
            )

        # Filter for input parameters which are provided in the transaction
        # Optional fields in the transaction model are set to None
        params = [
            self.authorize,
            self.unauthorize,
            self.authorize_credentials,
            self.unauthorize_credentials,
        ]
        if sum(param is not None for param in params) > 1:
            errors["DepositPreauth"] = (
                "DepositPreauth txn accepts exactly one input amongst `authorize`, "
                + "`unauthorize`, `authorize_credentials` and `unauthorize_credentials`"
                + "."
            )

        def _validate_credentials_length(
            credentials: List[Credential], field_name: str
        ) -> None:
            if credentials is None:
                return

            if len(credentials) == 0:
                errors["DepositPreauth"] = f"{field_name} list cannot be empty. "
            elif len(credentials) > MAX_CREDENTIAL_ARRAY_LENGTH:
                errors["DepositPreauth"] = (
                    f"{field_name} list cannot exceed "
                    + str(MAX_CREDENTIAL_ARRAY_LENGTH)
                    + " elements. "
                )

            if len(credentials) != len(set(credentials)):
                errors["DepositPreauth"] = (
                    f"{field_name} list cannot contain duplicate credentials."
                )

        if self.authorize_credentials is not None:
            _validate_credentials_length(
                self.authorize_credentials, "AuthorizeCredentials"
            )

        if self.unauthorize_credentials is not None:
            _validate_credentials_length(
                self.unauthorize_credentials, "UnauthorizeCredentials"
            )

        return errors


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Credential(NestedModel):
    """
    An inner object representing individual element inside AuthorizeCredentials and
    UnauthorizeCredentials array.
    """

    issuer: str = REQUIRED  # type: ignore
    """The issuer of the credential."""

    credential_type: str = REQUIRED  # type: ignore
    """A hex-encoded value to identify the type of credential from the issuer."""



"""Specifies an MPT amount."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict

from typing_extensions import Self

from xrpl.models.base_model import BaseModel
from xrpl.models.currencies.mpt_currency import MPTCurrency
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class MPTAmount(BaseModel):
    """Specifies an MPT amount."""

    mpt_issuance_id: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    value: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    def to_dict(self: Self) -> Dict[str, str]:
        """
        Returns the dictionary representation of an MPTAmount.

        Returns:
            The dictionary representation of an MPTAmount.
        """
        return {**super().to_dict(), "value": str(self.value)}

    def to_currency(self: Self) -> MPTCurrency:
        """
        Build an MPTCurrency from this MPTAmount.

        Returns:
            The MPTCurrency for this MPTAmount.
        """
        return MPTCurrency(mpt_issuance_id=self.mpt_issuance_id)



"""Model for PermissionedDomainSet transaction type."""

from dataclasses import dataclass, field
from typing import Dict, List, Optional

from typing_extensions import Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.deposit_preauth import Credential
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init

_MAX_ACCEPTED_CREDENTIALS_LENGTH = 10


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class PermissionedDomainSet(Transaction):
    """This transaction creates or modifies a PermissionedDomain object."""

    domain_id: Optional[str] = None
    """The domain to modify. Must be included if modifying an existing domain."""

    accepted_credentials: List[Credential] = REQUIRED  # type: ignore
    """The credentials that are accepted by the domain. Ownership of one of these
    credentials automatically makes you a member of the domain. An empty array means
    deleting the field."""

    transaction_type: TransactionType = field(
        default=TransactionType.PERMISSIONED_DOMAIN_SET,
        init=False,
    )
    """The transaction type (PermissionedDomainSet)."""

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        if len(self.accepted_credentials) == 0:
            errors["PermissionedDomainSet"] = (
                "AcceptedCredentials list cannot be empty."
            )
        elif len(self.accepted_credentials) > _MAX_ACCEPTED_CREDENTIALS_LENGTH:
            errors["PermissionedDomainSet"] = (
                "AcceptedCredentials list cannot have more than "
                + f"{_MAX_ACCEPTED_CREDENTIALS_LENGTH} elements."
            )
        elif len(self.accepted_credentials) != len(set(self.accepted_credentials)):
            errors["PermissionedDomainSet"] = (
                "AcceptedCredentials list cannot contain duplicate credentials."
            )

        return errors



"""
This module defines the GetAggregatePrice request API. It is used to fetch aggregate
statistics about the specified PriceOracles
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

from typing_extensions import Self

from xrpl.models.requests.ledger_entry import Oracle
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class GetAggregatePrice(Request):
    """
    The get_aggregate_price method retrieves the aggregate price of specified Oracle
    objects, returning three price statistics: mean, median, and trimmed mean.
    """

    method: RequestMethod = field(default=RequestMethod.GET_AGGREGATE_PRICE, init=False)

    base_asset: str = REQUIRED  # type: ignore
    """The currency code of the asset to be priced"""

    quote_asset: str = REQUIRED  # type: ignore
    """The currency code of the asset to quote the price of the base asset"""

    oracles: List[Oracle] = REQUIRED  # type: ignore
    """The oracle identifier"""

    trim: Optional[int] = None
    """The percentage of outliers to trim. Valid trim range is 1-25. If included, the
    API returns statistics for the trimmed mean"""

    time_threshold: Optional[int] = None
    """Defines a time range in seconds for filtering out older price data. Default
    value is 0, which doesn't filter any data"""

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if len(self.oracles) == 0:
            errors["GetAggregatePrice"] = (
                "Oracles array must contain at least one element"
            )
        return errors



"""Model for OracleDelete transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class OracleDelete(Transaction):
    """Delete an Oracle ledger entry."""

    account: str = REQUIRED  # type: ignore
    """This account must match the account in the Owner field of the Oracle object."""

    oracle_document_id: int = REQUIRED  # type: ignore
    """A unique identifier of the price oracle for the Account."""

    transaction_type: TransactionType = field(
        default=TransactionType.ORACLE_DELETE,
        init=False,
    )



"""Handles wallet generation from a faucet."""

import asyncio
from typing import Dict, Optional
from urllib.parse import urlparse, urlunparse

import httpx
from typing_extensions import Final

from xrpl.asyncio.account import get_balance, get_next_valid_seq_number
from xrpl.asyncio.clients import Client, XRPLRequestFailureException
from xrpl.asyncio.clients.client import get_network_id_and_build_version
from xrpl.constants import XRPLException
from xrpl.wallet.main import Wallet

_TEST_FAUCET_URL: Final[str] = "https://faucet.altnet.rippletest.net/accounts"
_DEV_FAUCET_URL: Final[str] = "https://faucet.devnet.rippletest.net/accounts"

_TIMEOUT_SECONDS: Final[int] = 40
_NETWORK_ID_URL_MAP: Dict[int, str] = {1: _TEST_FAUCET_URL, 2: _DEV_FAUCET_URL}


class XRPLFaucetException(XRPLException):
    """Faucet generation exception."""

    pass


async def generate_faucet_wallet(
    client: Client,
    wallet: Optional[Wallet] = None,
    debug: bool = False,
    faucet_host: Optional[str] = None,
    usage_context: Optional[str] = None,
    user_agent: Optional[str] = "xrpl-py",
) -> Wallet:
    """
    Generates a random wallet and funds it using the XRPL Testnet Faucet.

    Args:
        client: the network client used to make network calls.
        wallet: the wallet to fund. If omitted or `None`, a new wallet is created.
        debug: Whether to print debug information as it creates the wallet.
        faucet_host: A custom host to use for funding a wallet. In environments other
            than devnet and testnet, this parameter is required.
        usage_context: The intended use case for the funding request
            (for example, testing). This information  will be included
            in the json body of the HTTP request to the faucet.
        user_agent: A string representing the user agent (software/ client used)
            for the HTTP request. Default is "xrpl-py".


    Returns:
        A Wallet on the testnet that contains some amount of XRP.

    Raises:
        XRPLFaucetException: if an address could not be funded with the faucet.
        XRPLRequestFailureException: if a request to the ledger fails.
        requests.exceptions.HTTPError: if the request to the faucet fails.

    .. # noqa: DAR402 exception raised in private method
    """
    await get_network_id_and_build_version(client)

    if faucet_host is not None:
        faucet_url = process_faucet_host_url(faucet_host)
    else:
        if client.network_id is None:
            raise XRPLFaucetException(
                "Cannot create faucet URL without network_id or the faucet_host "
                + "information"
            )
        faucet_url = get_faucet_url(client.network_id)

    if wallet is None:
        wallet = Wallet.create()

    address = wallet.address
    # The faucet *can* be flakey... by printing info about this it's easier to
    # understand if tests are actually failing, or if it was just a faucet failure.
    if debug:
        print("Attempting to fund address {}".format(address))
    # Balance prior to asking for more funds
    starting_balance = await _check_wallet_balance(address, client)

    # Ask the faucet to send funds to the given address
    await _request_funding(faucet_url, address, usage_context, user_agent)
    # Wait for the faucet to fund our account or until timeout
    # Waits one second checks if balance has changed
    # If balance doesn't change it will attempt again until _TIMEOUT_SECONDS
    is_funded = False
    for _ in range(_TIMEOUT_SECONDS):
        await asyncio.sleep(1)
        if not is_funded:  # faucet transaction hasn't been validated yet
            current_balance = await _check_wallet_balance(address, client)
            # If our current balance has changed, then the account has been funded
            if current_balance > starting_balance:
                if debug:
                    print("Faucet fund successful.")
                is_funded = True
        else:  # wallet has been funded, now the ledger needs to know the account exists
            next_seq_num = await _try_to_get_next_seq(address, client)
            if next_seq_num is not None:
                return wallet

    raise XRPLFaucetException(
        "Unable to fund address with faucet after waiting {} seconds".format(
            _TIMEOUT_SECONDS
        )
    )


def process_faucet_host_url(input_url: str) -> str:
    """
    Construct a URL from the given input string.

    Args:
        input_url (str): The input string that may or may not include a protocol,
                       and may or may not have a path.

    Returns:
        str: The constructed URL with https as the default protocol and /accounts as the
        default path.
    """
    # Strip the trailing forward slash
    input_url = input_url.rstrip("/")

    # prepend the layer-5 internet protocol, if not already present
    # Read the comment about netloc to understand the behavior of urllib.urlparse
    # without the protocol at the beginning of the URL
    if "://" not in input_url:
        input_url = "https://" + input_url

    # Parse the input URL to identify its components.
    parsed_url = urlparse(input_url)

    # If the input string includes a protocol (e.g., "https://"), urlparse will
    # correctly parse it.
    # Scheme refers to the protocol (e.g., "https", "http").
    scheme = parsed_url.scheme if parsed_url.scheme else "https"

    # Netloc is the network location part, which usually includes the domain name.
    # For input "https://abcd.com", netloc is "abcd.com".
    # If no protocol is provided, the domain might be parsed as the path.
    # Consider the input string "abcd.com". If you were to parse this string using
    # urlparse without manually prepending a protocol (like http:// or https://), the
    # parsing logic would interpret "abcd.com" not as the network location part
    # (or domain) of the URL, but rather as the path component. This is because
    # urlparse expects a scheme (protocol) to correctly identify the parts of the URL.
    # Hence, we check if netloc is present; if not, assume the path is actually the
    # netloc.
    netloc = parsed_url.netloc if parsed_url.netloc else parsed_url.path
    path = parsed_url.path if parsed_url.netloc else ""

    # If no specific path is provided, append '/accounts' to the URL.
    # For input "abcd.com", the constructed path will be "/accounts".
    if not path:
        path = "/accounts"

    # Construct the final URL by reassembling its components.
    final_url = urlunparse((scheme, netloc, path, "", "", ""))

    return final_url


def get_faucet_url(network_id: int) -> str:
    """
    Returns the URL of the faucet that should be used, based on network_id

    Args:
        network_id: The network_id corresponding to the XRPL network. This is parsed
            from a ServerInfo() rippled response

    Returns:
        The URL of the matching faucet.

    Raises:
        XRPLFaucetException: if the provided network_id does not correspond to testnet
            or devnet.
    """
    if network_id in _NETWORK_ID_URL_MAP:
        return _NETWORK_ID_URL_MAP[network_id]

    # corresponds to sidechain-net2 network
    if network_id == 262:
        raise XRPLFaucetException(
            "Cannot fund an account on an issuing chain. Accounts must be created via "
            "the bridge."
        )
    elif network_id == 0:
        raise XRPLFaucetException("Cannot create faucet with a client on mainnet.")
    elif network_id < 0:
        raise XRPLFaucetException("Unable to parse the specified network_id.")

    # this line is unreachable. Custom devnets must specify a faucet_host input
    raise XRPLFaucetException(
        "The NetworkID of the provided network ( "
        + str(network_id)
        + ") does not have a known faucet."
    )


async def _check_wallet_balance(address: str, client: Client) -> int:
    try:
        return await get_balance(address, client)
    except XRPLRequestFailureException as e:
        if e.error == "actNotFound":  # transaction has not gone through
            return 0
        # some other error
        raise


async def _request_funding(
    url: str,
    address: str,
    usage_context: Optional[str] = None,
    user_agent: Optional[str] = None,
) -> None:
    async with httpx.AsyncClient() as http_client:
        json_body = {"destination": address, "userAgent": user_agent}
        if usage_context is not None:
            json_body["usageContext"] = usage_context
        response = await http_client.post(url=url, json=json_body)
    if not response.status_code == httpx.codes.OK:
        response.raise_for_status()


async def _try_to_get_next_seq(address: str, client: Client) -> Optional[int]:
    try:
        return await get_next_valid_seq_number(address, client)
    except XRPLRequestFailureException as e:
        if e.error == "actNotFound":
            # faucet gen has not fully gone through, try again
            return None
        # some other error
        raise



"""This request simulates a transaction without submitting it to the network."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Type

from typing_extensions import Self

from xrpl.models.exceptions import XRPLModelException
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Simulate(Request):
    """
    The `simulate` method simulates a transaction without submitting it to the
    network.
    """

    """
    The transaction to simulate, in
    `binary format <https://xrpl.org/docs/references/protocol/binary-format>`_. If you
    include this field, do not also include `transaction`.
    """
    tx_blob: Optional[str] = None

    """
    The transaction to simulate, in JSON format. If you include this field, do not also
    include `tx_blob`.
    """
    transaction: Optional[Transaction] = None

    """
    The default value is `false`, which returns data and metadata in JSON format. If
    `true`, returns data and metadata in binary format, serialized to a hexadecimal
    string.
    """
    binary: Optional[bool] = None

    method: RequestMethod = field(default=RequestMethod.SIMULATE, init=False)

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        # Can't have both `tx_blob` and `transaction`
        # Can't have neither `tx_blob` nor `transaction`
        if (self.tx_blob is None) == (self.transaction is None):
            errors["tx"] = (
                "Must have exactly one of `tx_blob` and `transaction` fields."
            )

        if self.transaction is not None and self.transaction.is_signed():
            errors["transaction"] = "Cannot simulate a signed transaction."

        return errors

    @classmethod
    def from_dict(cls: Type[Self], value: Dict[str, Any]) -> Self:
        """
        Construct a new Simulate from a dictionary of parameters.

        Args:
            value: The value to construct the Simulate from.

        Returns:
            A new Simulate object, constructed using the given parameters.

        Raises:
            XRPLModelException: If the `tx_json` value provided is not a valid
                transaction.
        """
        if "tx_json" in value:
            try:
                fixed_value = {
                    **value,
                    "transaction": Transaction.from_xrpl(value["tx_json"]),
                }
            except XRPLModelException as e:
                raise XRPLModelException(f"Invalid tx_json format: {str(e)}") from e
            del fixed_value["tx_json"]
        else:
            fixed_value = value
        return super().from_dict(fixed_value)

    def to_dict(self: Self) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a Simulate.

        Returns:
            The dictionary representation of a Simulate.
        """
        return_dict = super().to_dict()
        if self.transaction is not None:
            del return_dict["transaction"]
            return_dict["tx_json"] = self.transaction.to_xrpl()
        return return_dict



"""Model for MPTokenIssuanceDestroy transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class MPTokenIssuanceDestroy(Transaction):
    """
    The MPTokenIssuanceDestroy transaction is used to remove an MPTokenIssuance object
    from the directory node in which it is being held, effectively removing the token
    from the ledger. If this operation succeeds, the corresponding
    MPTokenIssuance is removed and the owner’s reserve requirement is reduced by one.
    This operation must fail if there are any holders who have non-zero balances.
    """

    mptoken_issuance_id: str = REQUIRED  # type: ignore
    """Identifies the MPTokenIssuance object to be removed by the transaction."""

    transaction_type: TransactionType = field(
        default=TransactionType.MPTOKEN_ISSUANCE_DESTROY,
        init=False,
    )



"""Model for MPTokenIssuanceCreate transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from typing_extensions import Final, Self

from xrpl.constants import HEX_REGEX
from xrpl.models.transactions.transaction import Transaction, TransactionFlagInterface
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init

_MAX_TRANSFER_FEE: Final[int] = 50000


class MPTokenIssuanceCreateFlag(int, Enum):
    """
    Transactions of the MPTokenIssuanceCreate type support additional values in the
    Flags field.
    This enum represents those options.
    """

    TF_MPT_CAN_LOCK = 0x00000002
    TF_MPT_REQUIRE_AUTH = 0x00000004
    TF_MPT_CAN_ESCROW = 0x00000008
    TF_MPT_CAN_TRADE = 0x00000010
    TF_MPT_CAN_TRANSFER = 0x00000020
    TF_MPT_CAN_CLAWBACK = 0x00000040


class MPTokenIssuanceCreateFlagInterface(TransactionFlagInterface):
    """
    Transactions of the MPTokenIssuanceCreate type support additional values in the
    Flags field.
    This TypedDict represents those options.
    """

    TF_MPT_CAN_LOCK: bool
    TF_MPT_REQUIRE_AUTH: bool
    TF_MPT_CAN_ESCROW: bool
    TF_MPT_CAN_TRADE: bool
    TF_MPT_CAN_TRANSFER: bool
    TF_MPT_CAN_CLAWBACK: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class MPTokenIssuanceCreate(Transaction):
    """
    The MPTokenIssuanceCreate transaction creates a MPTokenIssuance object
    and adds it to the relevant directory node of the creator account.
    This transaction is the only opportunity an issuer has to specify any token fields
    that are defined as immutable (e.g., MPT Flags). If the transaction is successful,
    the newly created token will be owned by the account (the creator account) which
    executed the transaction.
    """

    asset_scale: Optional[int] = None
    """
    An asset scale is the difference, in orders of magnitude, between a standard unit
    and a corresponding fractional unit. More formally, the asset scale is a
    non-negative integer (0, 1, 2, …) such that one standard unit equals 10^(-scale) of
    a corresponding fractional unit. If the fractional unit equals the standard unit,
    then the asset scale is 0.
    Note that this value is optional, and will default to 0 if not supplied.
    """

    maximum_amount: Optional[str] = None
    """
    Specifies the maximum asset amount of this token that should ever be issued.
    It is a non-negative integer string that can store a range of up to 63 bits. If
    not set, the max amount will default to the largest unsigned 63-bit integer
    (0x7FFFFFFFFFFFFFFF)
    """

    transfer_fee: Optional[int] = None
    """
    Specifies the fee to charged by the issuer for secondary sales of the Token,
    if such sales are allowed. Valid values for this field are between 0 and 50,000
    inclusive, allowing transfer rates of between 0.000% and 50.000% in increments of
    0.001. The field must NOT be present if the `tfMPTCanTransfer` flag is not set.
    """

    mptoken_metadata: Optional[str] = None
    """
    Arbitrary metadata about this issuance, in hex format.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.MPTOKEN_ISSUANCE_CREATE,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        if self.transfer_fee is not None:
            if not self.has_flag(MPTokenIssuanceCreateFlag.TF_MPT_CAN_TRANSFER):
                errors["transfer_fee"] = (
                    "Field cannot be provided without enabling tfMPTCanTransfer flag."
                )
            if self.transfer_fee < 0 or self.transfer_fee > _MAX_TRANSFER_FEE:
                errors["transfer_fee"] = "Field must be between 0 and " + str(
                    _MAX_TRANSFER_FEE
                )

        if self.mptoken_metadata is not None:
            if len(self.mptoken_metadata) == 0:
                errors["mptoken_metadata"] = "Field must not be empty string."
            elif bool(HEX_REGEX.fullmatch(self.mptoken_metadata)) is False:
                errors["mptoken_metadata"] = "Field must be in hex format."

        return errors



"""
Class for serializing and deserializing a 64-bit UInt.
See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
"""

from __future__ import annotations

import re
from typing import Optional, Pattern, Type, Union

from typing_extensions import Final, Self

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.uint import UInt

_WIDTH: Final[int] = 8  # 64 / 8

_BASE10_REGEX: Final[Pattern[str]] = re.compile("^[0-9]{1,20}$")
_HEX_REGEX: Final[Pattern[str]] = re.compile("^[a-fA-F0-9]{1,16}$")

SPECIAL_FIELDS: Final[set[str]] = {
    "MaximumAmount",
    "OutstandingAmount",
    "MPTAmount",
}


class UInt64(UInt):
    """
    Class for serializing and deserializing a 64-bit UInt.
    See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
    """

    def __init__(self: Self, buffer: bytes = bytes(_WIDTH)) -> None:
        """Construct a new UInt64 type from a ``bytes`` value."""
        super().__init__(buffer)

    @classmethod
    def from_parser(
        cls: Type[Self], parser: BinaryParser, _length_hint: Optional[int] = None
    ) -> Self:
        """
        Construct a new UInt64 type from a BinaryParser.

        Args:
            parser: The BinaryParser to construct a UInt64 from.

        Returns:
            The UInt64 constructed from parser.
        """
        return cls(parser.read(_WIDTH))

    @classmethod
    def from_value(
        cls: Type[Self], value: Union[str, int], field_name: str = ""
    ) -> Self:
        """
        Construct a new UInt64 type from a value.

        Args:
            value: The value to construct a UInt64 from.
            field_name: The optional field name (for special handling
                        of base 10 strings).

        Returns:
            The UInt64 constructed from the value.

        Raises:
            XRPLBinaryCodecException: If a UInt64 could not be constructed
                                      from the value.
        """
        if isinstance(value, int):
            if value < 0:
                raise XRPLBinaryCodecException(f"{value} must be an unsigned integer")
            value_bytes = value.to_bytes(_WIDTH, byteorder="big", signed=False)
            return cls(value_bytes)

        if isinstance(value, str):
            if field_name in SPECIAL_FIELDS and _BASE10_REGEX.fullmatch(value):
                # Convert base 10 string to hex string
                value = hex(int(value))[2:]

            if not _HEX_REGEX.fullmatch(value):
                raise XRPLBinaryCodecException(f"{value} is not a valid hex string")

            value = value.rjust(16, "0")
            value_bytes = bytes.fromhex(value)
            return cls(value_bytes)

        raise XRPLBinaryCodecException(
            f"Cannot construct UInt64 from given value {value}"
        )

    def to_json(self: Self, field_name: str = "") -> str:
        """
        Convert a UInt64 object to JSON (hex or base 10, depending on field_name).

        Args:
            field_name: The optional field name (for special handling
                        of base 10 format).

        Returns:
            The JSON representation of the UInt64 object.
        """
        hex_string = self.buffer.hex().upper()
        if field_name in SPECIAL_FIELDS:
            return str(int(hex_string, 16))  # Return base 10 string
        return hex_string



"""Model for MPTokenIssuanceSet transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction, TransactionFlagInterface
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


class MPTokenIssuanceSetFlag(int, Enum):
    """
    Transactions of the MPTokenIssuanceSet type support additional values in the
    Flags field.
    This enum represents those options.
    """

    TF_MPT_LOCK = 0x00000001
    """
    If set, indicates that the MPT can be locked both individually and globally.
    If not set, the MPT cannot be locked in any way.
    """

    TF_MPT_UNLOCK = 0x00000002
    """
    If set, indicates that the MPT can be unlocked both individually and globally.
    If not set, the MPT cannot be unlocked in any way.
    """


class MPTokenIssuanceSetFlagInterface(TransactionFlagInterface):
    """
    Transactions of the MPTokenIssuanceSet type support additional values in the
    Flags field.
    This TypedDict represents those options.
    """

    TF_MPT_LOCK: bool
    TF_MPT_UNLOCK: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class MPTokenIssuanceSet(Transaction):
    """
    The MPTokenIssuanceSet transaction is used to globally lock/unlock a
    MPTokenIssuance, or lock/unlock an individual's MPToken.
    """

    mptoken_issuance_id: str = REQUIRED  # type: ignore
    """Identifies the MPTokenIssuance"""

    holder: Optional[str] = None
    """
    An optional XRPL Address of an individual token holder balance to lock/unlock.
    If omitted, this transaction will apply to all any accounts holding MPTs.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.MPTOKEN_ISSUANCE_SET,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        if self.has_flag(MPTokenIssuanceSetFlag.TF_MPT_LOCK) and self.has_flag(
            MPTokenIssuanceSetFlag.TF_MPT_UNLOCK
        ):
            errors["flags"] = (
                "flag conflict: both TF_MPT_LOCK and TF_MPT_UNLOCK can't be set"
            )

        return errors



"""
The server_state command asks the server for various
machine-readable information about the rippled server's
current state. The response is almost the same as the
server_info method, but uses units that are easier to
process instead of easier to read. (For example, XRP
values are given in integer drops instead of scientific
notation or decimal values, and time is given in
milliseconds instead of seconds.)
"""

from dataclasses import dataclass, field

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class ServerState(Request):
    """
    The server_state command asks the server for various
    machine-readable information about the rippled server's
    current state. The response is almost the same as the
    server_info method, but uses units that are easier to
    process instead of easier to read. (For example, XRP
    values are given in integer drops instead of scientific
    notation or decimal values, and time is given in
    milliseconds instead of seconds.)
    """

    method: RequestMethod = field(default=RequestMethod.SERVER_STATE, init=False)



"""High-level reliable submission methods with XRPL transactions."""

import asyncio
from typing import Optional

from xrpl.asyncio.transaction import submit_and_wait as async_submit_and_wait
from xrpl.clients.sync_client import SyncClient
from xrpl.models.response import Response
from xrpl.models.transactions.transaction import Transaction
from xrpl.wallet.main import Wallet


def submit_and_wait(
    transaction: Transaction,
    client: SyncClient,
    wallet: Optional[Wallet] = None,
    *,
    check_fee: bool = True,
    autofill: bool = True,
    fail_hard: bool = False,
) -> Response:
    """
    Signs a transaction locally, without trusting external rippled nodes (only if
    the input transaction is unsigned; otherwise, proceeds to the next steps), submits,
    and verifies that it has been included in a validated ledger (or has errored
    /will not be included for some reason).
    `See Reliable Transaction Submission
    <https://xrpl.org/reliable-transaction-submission.html>`_

    Args:
        transaction: the signed/unsigned transaction (or transaction blob) to
            be submitted.
        client: the network client with which to submit the transaction.
        wallet: an optional wallet with which to sign the transaction. This is
            only needed if the transaction is unsigned.
        check_fee: an optional bolean indicating whether to check if the fee is
            higher than the expected transaction type fee. Defaults to True.
        autofill: an optional boolean indicating whether to autofill the
            transaction. Defaults to True.
        fail_hard: an optional boolean. If True, and the transaction fails for
            the initial server, do not retry or relay the transaction to other
            servers. Defaults to False.

    Returns:
        The response from the ledger.
    """
    return asyncio.run(
        async_submit_and_wait(
            transaction,
            client,
            wallet,
            check_fee=check_fee,
            autofill=autofill,
            fail_hard=fail_hard,
        )
    )



"""Helper util functions for the models module."""

import re
from dataclasses import dataclass, is_dataclass
from typing import Any, Dict, List, Optional, Pattern, Type, TypeVar, cast

from typing_extensions import Final

from xrpl.models.exceptions import XRPLModelException

HEX_REGEX: Final[Pattern[str]] = re.compile("[a-fA-F0-9]*")

MAX_CREDENTIAL_ARRAY_LENGTH = 8

# Credentials are represented in hex. Whilst they are allowed a maximum length of 64
# bytes, every byte requires 2 hex characters for representation
_MAX_CREDENTIAL_LENGTH: Final[int] = 128


def get_credential_type_error(credential_type: str) -> Optional[str]:
    """
    Utility function for validating the CredentialType field in all
    transactions related to Credential.

    Args:
        credential_type: A hex-encoded value to identify the type of credential from
            the issuer.

    Returns:
        Errors, if any, during validation of credential_type field
    """
    errors = []
    # credential_type is a required field in this transaction
    if len(credential_type) == 0:
        errors.append("cannot be an empty string.")
    elif len(credential_type) > _MAX_CREDENTIAL_LENGTH:
        errors.append(f"Length cannot exceed {_MAX_CREDENTIAL_LENGTH}.")
    if not HEX_REGEX.fullmatch(credential_type):
        errors.append("credential_type field must be encoded in hex.")
    return " ".join(errors) if len(errors) > 0 else None


def validate_credential_ids(credential_list: Optional[List[str]]) -> Dict[str, str]:
    """
    Args:
        credential_list: An optional list of input credentials

    Returns:
        Errors pertaining to credential_ids field
    """
    errors: Dict[str, str] = {}
    if credential_list is None:
        return errors

    if len(credential_list) == 0:
        errors["credential_ids"] = "CredentialIDs list cannot be empty."
    elif len(credential_list) > MAX_CREDENTIAL_ARRAY_LENGTH:
        errors["credential_ids"] = (
            f"CredentialIDs list cannot exceed {MAX_CREDENTIAL_ARRAY_LENGTH}"
            + " elements."
        )

    if len(credential_list) != len(set(credential_list)):
        errors["credential_ids_duplicates"] = (
            "CredentialIDs list cannot contain duplicate values."
        )

    return errors


# Code source for requiring kwargs:
# https://gist.github.com/mikeholler/4be180627d3f8fceb55704b729464adb

_T = TypeVar("_T")
_Self = TypeVar("_Self")


def _is_kw_only_attr_defined_in_dataclass() -> bool:
    """
    Returns:
        Utility function to determine if the Python interpreter's version is older
        than 3.10. This information is used to check the presence of KW_ONLY attribute
        in the dataclass

    For ease of understanding, the output of this function should be equivalent to the
    below code, unless the `kw_only` attribute is backported to older versions of
    Python interpreter

    Returns:
    if sys.version_info.major < 3:
        return True
    return sys.version_info.minor < 10
    """
    return "kw_only" in dataclass.__kwdefaults__


# Python 3.10 and higer versions of Python enable a new KW_ONLY parameter in dataclass
# This dictionary is used to ensure that Ledger Objects constructors reject
# positional arguments. It obviates the need to maintain decorators for the same
# functionality and enbles IDEs to auto-complete the constructor arguments.
# KW_ONLY Docs: https://docs.python.org/3/library/dataclasses.html#dataclasses.KW_ONLY

# Unit tests that validate this behavior can be found at test_channel_authorize.py
# and test_sign.py files.
KW_ONLY_DATACLASS = (
    dict(kw_only=True) if _is_kw_only_attr_defined_in_dataclass() else {}
)


def require_kwargs_on_init(cls: Type[_T]) -> Type[_T]:
    """
    Force a dataclass's init function to only work if called with keyword arguments.
    If parameters are not positional-only, a TypeError is thrown with a helpful message.
    This function may only be used on dataclasses.

    This works by wrapping the __init__ function and dynamically replacing it.
    Therefore, stacktraces for calls to the new __init__ might look a bit strange. Fear
    not though, all is well.

    Note: although this may be used as a decorator, this is not advised as IDEs will no
    longer suggest parameters in the constructor. Instead, this is the recommended
    usage:
        from dataclasses import dataclass
        @dataclass
        class Foo:
            bar: str
        require_kwargs_on_init(Foo)

    Args:
        cls: The class that requires keyword arguments (must be a dataclass).

    Returns:
        The provided class, adding an error on init if positional args are provided.

    Raises:
        TypeError: If cls is None or is not a dataclass.
    """
    # error messages for dev help
    if cls is None:
        raise TypeError("Cannot call with cls=None")
    if not is_dataclass(cls):
        raise TypeError(
            f"This decorator only works on dataclasses. {cls.__name__} is not a "
            "dataclass."
        )

    original_init = cls.__init__

    def new_init(self: _Self, *args: List[Any], **kwargs: Dict[str, Any]) -> None:
        if len(args) > 0:
            raise XRPLModelException(
                f"{type(self).__name__}.__init__ only allows keyword arguments. "
                f"Found the following positional arguments: {args}"
            )
        original_init(self, **kwargs)

    # For Python v3.10 and above, the KW_ONLY attribute in data_class
    # performs the functionality of require_kwargs_on_init class.
    # When support for older versions of Python (earlier than v3.10) is removed, the
    # usage of require_kwargs_on_init decorator on model classes can also be removed.
    if not _is_kw_only_attr_defined_in_dataclass():
        cls.__init__ = new_init  # type: ignore
    return cast(Type[_T], cls)



"""Model for DelegateSet transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Union

from typing_extensions import Self

from xrpl.models.nested_model import NestedModel
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init

PERMISSIONS_MAX_LENGTH = 10

NON_DELEGABLE_TRANSACTIONS = {
    TransactionType.ACCOUNT_SET,
    TransactionType.SET_REGULAR_KEY,
    TransactionType.SIGNER_LIST_SET,
    TransactionType.DELEGATE_SET,
    TransactionType.ACCOUNT_DELETE,
    TransactionType.BATCH,
}


class GranularPermission(str, Enum):
    """
    These permissions would support control over some smaller portion of a transaction,
    rather than being able to do all of the functionality that the transaction allows.
    """

    TRUSTLINE_AUTHORIZE = "TrustlineAuthorize"
    """Authorize a trustline."""

    TRUSTLINE_FREEZE = "TrustlineFreeze"
    """Freeze a trustline."""

    TRUSTLINE_UNFREEZE = "TrustlineUnfreeze"
    """Unfreeze a trustline."""

    ACCOUNT_DOMAIN_SET = "AccountDomainSet"
    """Modify the domain of an account."""

    ACCOUNT_EMAIL_HASH_SET = "AccountEmailHashSet"
    """Modify the EmailHash of an account."""

    ACCOUNT_MESSAGE_KEY_SET = "AccountMessageKeySet"
    """Modify the MessageKey of an account."""

    ACCOUNT_TRANSFER_RATE_SET = "AccountTransferRateSet"
    """Modify the transfer rate of an account."""

    ACCOUNT_TICK_SIZE_SET = "AccountTickSizeSet"
    """Modify the tick size of an account."""

    PAYMENT_MINT = "PaymentMint"
    """Send a payment for a currency where the sending account is the issuer."""

    PAYMENT_BURN = "PaymentBurn"
    """Send a payment for a currency where the destination account is the issuer."""

    MPTOKEN_ISSUANCE_LOCK = "MPTokenIssuanceLock"
    """Use the MPTIssuanceSet transaction to lock (freeze) a holder."""

    MPTOKEN_ISSUANCE_UNLOCK = "MPTokenIssuanceUnlock"
    """Use the MPTIssuanceSet transaction to unlock (unfreeze) a holder."""


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Permission(NestedModel):
    """Represents one entry in a Permissions list used in DelegateSet
    transaction.
    """

    permission_value: Union[
        TransactionType, GranularPermission
    ] = REQUIRED  # type: ignore
    """
    Transaction level or granular permission.

    :meta hide-value:
    """


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class DelegateSet(Transaction):
    """DelegateSet allows an account to delegate a set of permissions to another
    account.
    """

    authorize: str = REQUIRED  # type: ignore
    """The authorized account."""

    permissions: List[Permission] = REQUIRED  # type: ignore
    """The transaction permissions that the authorized account has been granted."""

    transaction_type: TransactionType = field(
        default=TransactionType.DELEGATE_SET,
        init=False,
    )
    """The transaction type (DelegateSet)."""

    def _get_errors(self: Self) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "permissions": self._get_permissions_error(),
                "account_addresses": self._validate_account_addresses(),
            }.items()
            if value is not None
        }

    def _validate_account_addresses(self: Self) -> Optional[str]:
        if self.authorize == self.account:
            return "Field `authorize` and `account` must be different."
        return None

    def _get_permissions_error(self: Self) -> Optional[str]:
        if len(self.permissions) > PERMISSIONS_MAX_LENGTH:
            return (
                f"Length of `permissions` list is greater than "
                f"{PERMISSIONS_MAX_LENGTH}."
            )

        entered_permissions = [
            permission.permission_value for permission in self.permissions
        ]
        if len(entered_permissions) != len(set(entered_permissions)):
            return "Duplicate permission value in `permissions` list."

        if set(entered_permissions) & NON_DELEGABLE_TRANSACTIONS:
            return (
                f"Non-delegable transactions found in `permissions` list: "
                f"{set(entered_permissions) & NON_DELEGABLE_TRANSACTIONS}."
            )

        return None



"""Helper functions for parsers."""

from decimal import Decimal
from typing import Any, Dict, List, Union

from xrpl.utils.txn_parser.utils.types import (
    AccountBalance,
    AccountOfferChange,
    CurrencyAmount,
)


def get_value(balance: Union[CurrencyAmount, Dict[str, str], str]) -> Decimal:
    """
    Get a currency amount's value.

    Args:
        balance: Account's balance.

    Returns:
        The currency amount's value.
    """
    if isinstance(balance, str):
        return Decimal(balance)
    return Decimal(balance["value"])


def group_by_account(
    account_objects: Union[List[AccountBalance], List[AccountOfferChange]],
) -> Dict[str, Any]:
    """
    Groups the account objects in one list for each account.

    Args:
        account_objects: All computed objects.

    Returns:
        The grouped computed objects.
    """
    grouped_objects: Dict[str, Any] = {}
    for object in account_objects:
        if object.get("account") is not None:
            account = str(object.get("account"))
        else:
            account = str(object.get("maker_account"))
        grouped_objects.setdefault(account, []).append(object)
    return grouped_objects



"""Model for AccountDelete transaction type."""

from dataclasses import dataclass, field
from typing import Dict, List, Optional

from typing_extensions import Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import (
    KW_ONLY_DATACLASS,
    require_kwargs_on_init,
    validate_credential_ids,
)


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class AccountDelete(Transaction):
    """
    Represents an `AccountDelete transaction
    <https://xrpl.org/accountdelete.html>`_, which deletes an account and any
    objects it owns in the XRP Ledger, if possible, sending the account's remaining
    XRP to a specified destination account.

    See `Deletion of Accounts
    <https://xrpl.org/accounts.html#deletion-of-accounts>`_ for the requirements to
    delete an account.
    """

    destination: str = REQUIRED  # type: ignore
    """
    The address of the account to which to send any remaining XRP.
    This field is required.

    :meta hide-value:
    """

    destination_tag: Optional[int] = None
    """
    The `destination tag
    <https://xrpl.org/source-and-destination-tags.html>`_ at the
    ``destination`` account where funds should be sent.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.ACCOUNT_DELETE,
        init=False,
    )

    credential_ids: Optional[List[str]] = None
    """Credentials associated with sender of this transaction. The credentials included
    must not be expired. The list must not be empty when specified and cannot contain
    more than 8 credentials."""

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        errors.update(validate_credential_ids(self.credential_ids))
        return errors



"""Types used by the parser."""

from typing import List

from typing_extensions import Literal, NotRequired, TypedDict


class Balance(TypedDict):
    """A account's balance model."""

    currency: str
    """The currency code."""

    value: str
    """The amount of the currency."""

    issuer: NotRequired[str]
    """The issuer of the currency. This value is optional."""


class AccountBalance(TypedDict):
    """A single account balance."""

    account: str
    """The affected account."""
    balance: Balance
    """The balance."""


class AccountBalances(TypedDict):
    """A model representing an account's balances."""

    account: str
    balances: List[Balance]


class CurrencyAmount(Balance):
    """A currency amount model. Has the same fields as `Balance`"""

    pass


class OfferChange(TypedDict):
    """A single offer change."""

    flags: int
    taker_gets: CurrencyAmount
    taker_pays: CurrencyAmount
    sequence: int
    status: Literal["created", "partially-filled", "filled", "cancelled"]
    maker_exchange_rate: str
    expiration_time: NotRequired[int]


class AccountOfferChange(TypedDict):
    """A model representing an account's offer change."""

    maker_account: str
    offer_change: OfferChange


class AccountOfferChanges(TypedDict):
    """A model representing an account's offer changes."""

    maker_account: str
    offer_changes: List[OfferChange]



"""Helper functions for signing multi-account Batch transactions."""

from typing import List, Union, cast

from xrpl.constants import XRPLException
from xrpl.core.addresscodec.codec import decode_classic_address
from xrpl.core.binarycodec import encode, encode_for_signing_batch
from xrpl.core.binarycodec.main import BatchSigningDict
from xrpl.core.keypairs import sign
from xrpl.models.transactions import Batch, Signer, Transaction
from xrpl.models.transactions.batch import BatchSigner
from xrpl.wallet import Wallet


def sign_multiaccount_batch(
    wallet: Wallet, transaction: Batch, multisign: bool = False
) -> Batch:
    """
    Sign a multi-account Batch transaction.

    Args:
        wallet: Wallet instance.
        transaction: The Batch transaction to sign.
        multisign: Specify true/false to use multisign. Defaults to False.

    Raises:
        XRPLException: If the wallet signing the transaction doesn't have an account in
            the Batch.

    Returns:
        The Batch transaction with the BatchSigner included.
    """
    involved_accounts = set(tx.account for tx in transaction.raw_transactions)
    if wallet.address not in involved_accounts:
        raise XRPLException("Must be signing for an address included in the Batch.")

    fields_to_sign: BatchSigningDict = {
        "flags": transaction._flags_to_int() or 0,
        "transaction_ids": [tx.get_hash() for tx in transaction.raw_transactions],
    }
    if multisign:
        signer = Signer(
            account=wallet.address,
            signing_pub_key=wallet.public_key,
            txn_signature=sign(
                encode_for_signing_batch(fields_to_sign), wallet.private_key
            ),
        )
        batch_signer = BatchSigner(account=wallet.address, signers=[signer])
    else:
        batch_signer = BatchSigner(
            account=wallet.address,
            signing_pub_key=wallet.public_key,
            txn_signature=sign(
                encode_for_signing_batch(fields_to_sign), wallet.private_key
            ),
        )

    transaction_dict = transaction.to_dict()
    transaction_dict["batch_signers"] = [batch_signer]

    return Batch.from_dict(transaction_dict)


def combine_batch_signers(transactions: List[Union[Batch, str]]) -> str:
    """
    Takes several transactions with BatchSigners fields (in object or blob form) and
    creates a single transaction with all BatchSigners that then gets signed and
    returned.

    Args:
        transactions: The transactions to combine `BatchSigners` values on.

    Raises:
        XRPLException: If the list of transactions provided is invalid.

    Returns:
        A single signed Transaction which has all BatchSigners from transactions within
        it.
    """
    if len(transactions) == 0:
        raise XRPLException("There were 0 transactions to combine.")

    decoded_txs: List[Transaction] = [
        Transaction.from_blob(tx) if isinstance(tx, str) else tx for tx in transactions
    ]
    for tx in decoded_txs:
        if tx.transaction_type != "Batch":
            raise XRPLException("TransactionType must be `Batch`.")
        batch = cast(Batch, tx)
        if batch.batch_signers is None or len(batch.batch_signers) == 0:
            raise XRPLException(
                "For combining Batch transaction signatures, all transactions must "
                "include a BatchSigners field containing an array of signatures."
            )
        if (
            tx.signing_pub_key != ""
            or tx.txn_signature is not None
            or tx.signers is not None
        ):
            raise XRPLException("Transaction must be unsigned.")

    batch_txs = cast(List[Batch], decoded_txs)
    _validate_batch_equivalence(batch_txs)

    return encode(_get_batch_with_all_signers(batch_txs).to_xrpl())


def _validate_batch_equivalence(transactions: List[Batch]) -> None:
    example_tx = transactions[0]
    for tx in transactions:
        if (
            tx.flags != example_tx.flags
            or tx.raw_transactions != example_tx.raw_transactions
        ):
            raise XRPLException(
                "Flags and RawTransactions are not the same for all provided "
                "transactions."
            )


def _get_batch_with_all_signers(transactions: List[Batch]) -> Batch:
    batch_signers = [
        signer
        for tx in transactions
        if tx.batch_signers is not None
        for signer in tx.batch_signers
        if signer.account != transactions[0].account
    ]
    batch_signers.sort(
        key=lambda signer: decode_classic_address(signer.account).hex().upper()
    )
    returned_tx_dict = transactions[0].to_dict()
    return Batch.from_dict({**returned_tx_dict, "batch_signers": batch_signers})



"""The base model for all transactions and their nested object types."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from hashlib import sha512
from typing import Any, Dict, List, Optional, Type, Union

from typing_extensions import Final, Self

from xrpl.core.binarycodec import decode, encode
from xrpl.models.amounts import IssuedCurrencyAmount
from xrpl.models.amounts.mpt_amount import MPTAmount
from xrpl.models.base_model import ABBREVIATIONS, BaseModel
from xrpl.models.exceptions import XRPLModelException
from xrpl.models.flags import (
    FlagInterface,
    check_false_flag_definition,
    interface_to_flag_list,
)
from xrpl.models.nested_model import NestedModel
from xrpl.models.requests import PathStep
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.types import PseudoTransactionType, TransactionType
from xrpl.models.types import XRPL_VALUE_TYPE
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init

_TRANSACTION_HASH_PREFIX: Final[int] = 0x54584E00


def transaction_json_to_binary_codec_form(
    dictionary: Dict[str, XRPL_VALUE_TYPE],
) -> Dict[str, XRPL_VALUE_TYPE]:
    """
    Returns a new dictionary in which the keys have been formatted as CamelCase and
    standardized to be serialized by the binary codec.

    Args:
        dictionary: The dictionary to be reformatted.

    Returns:
        A new dictionary object that has been reformatted.
    """
    # This method should be made private when it is removed from `xrpl.transactions`
    return {
        _key_to_tx_json(key): _value_to_tx_json(value)
        for (key, value) in dictionary.items()
    }


def _key_to_tx_json(key: str) -> str:
    """
    Transforms snake_case to PascalCase. For example:
        1. 'transaction_type' becomes 'TransactionType'
        2. 'URI' becomes 'uri'

    Known abbreviations (example 2 above) need to be enumerated in ABBREVIATIONS.
    """
    return "".join(
        [
            ABBREVIATIONS[word] if word in ABBREVIATIONS else word.capitalize()
            for word in key.split("_")
        ]
    )


def _value_to_tx_json(value: XRPL_VALUE_TYPE) -> XRPL_VALUE_TYPE:
    # IssuedCurrencyAmount and PathStep are special cases and should not be snake cased
    # and only contain primitive members
    if isinstance(value, list) and all(PathStep.is_dict_of_model(v) for v in value):
        return value
    if IssuedCurrencyAmount.is_dict_of_model(value):
        return value
    if MPTAmount.is_dict_of_model(value):
        return value
    if isinstance(value, dict):
        return transaction_json_to_binary_codec_form(value)
    if isinstance(value, list):
        return [_value_to_tx_json(sub_value) for sub_value in value]
    return value


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Memo(NestedModel):
    """
    An arbitrary piece of data attached to a transaction. A transaction can
    have multiple Memo objects as an array in the Memos field.
    Must contain one or more of ``memo_data``, ``memo_format``, and
    ``memo_type``.
    """

    memo_data: Optional[str] = None
    """The data of the memo, as a hexadecimal string."""

    memo_format: Optional[str] = None
    """
    The format of the memo, as a hexadecimal string. Conventionally, this
    should be the `MIME type
    <http://www.iana.org/assignments/media-types/media-types.xhtml>`_
    of the memo data.
    """

    memo_type: Optional[str] = None
    """
    The type of the memo, as a hexadecimal string. Conventionally, this
    should be an `RFC 5988 relation
    <http://tools.ietf.org/html/rfc5988#section-4>`_ defining the format of
    the memo data.
    """

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        present_memo_fields = [
            field
            for field in [
                self.memo_data,
                self.memo_format,
                self.memo_type,
            ]
            if field is not None
        ]
        if len(present_memo_fields) < 1:
            errors["Memo"] = "Memo must contain at least one field"
        return errors


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Signer(NestedModel):
    """
    One Signer in a multi-signature. A multi-signed transaction can have an
    array of up to 8 Signers, each contributing a signature, in the Signers
    field.
    """

    account: str = REQUIRED  # type: ignore
    """
    The address of the Signer. This can be a funded account in the XRP
    Ledger or an unfunded address.
    This field is required.

    :meta hide-value:
    """

    txn_signature: str = REQUIRED  # type: ignore
    """
    The signature that this Signer provided for this transaction.
    This field is required.

    :meta hide-value:
    """

    signing_pub_key: str = REQUIRED  # type: ignore
    """
    The public key that should be used to verify this Signer's signature.
    This field is required.

    :meta hide-value:
    """


class TransactionFlag(int, Enum):
    """
    Transactions of the Transaction type support additional values in the Flags field.
    This enum represents those options.
    """

    TF_INNER_BATCH_TXN = 0x40000000


class TransactionFlagInterface(FlagInterface):
    """
    Transactions support additional values in the Flags field. This TypedDict
    represents those options.
    """

    TF_INNER_BATCH_TXN: bool


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Transaction(BaseModel):
    """
    The base class for all `transaction types
    <https://xrpl.org/transaction-types.html>`_. Represents `fields common to all
    transaction types <https://xrpl.org/transaction-common-fields.html>`_.
    """

    account: str = REQUIRED  # type: ignore
    """
    The address of the sender of the transaction. Required.

    :meta hide-value:
    """

    transaction_type: Union[
        TransactionType, PseudoTransactionType
    ] = REQUIRED  # type: ignore

    fee: Optional[str] = None  # auto-fillable
    """
    (Auto-fillable) The amount of XRP to destroy as a cost to send this
    transaction. See `Transaction Cost
    <https://xrpl.org/transaction-cost.html>`_ for details.
    """

    sequence: Optional[int] = None  # auto-fillable
    """
    (Auto-fillable) The sequence number of the transaction. Must match the
    sending account's next unused sequence number. See `Account Sequence
    <https://xrpl.org/basic-data-types.html#account-sequence>`_ for details.
    """

    account_txn_id: Optional[str] = None
    """
    A hash value identifying a previous transaction from the same sender. If
    provided, this transaction is only considered valid if the identified
    transaction is the most recent transaction sent by this address. See
    `AccountTxnID
    <https://xrpl.org/transaction-common-fields.html#accounttxnid>`_ for
    details.
    """

    flags: Optional[Union[Dict[str, bool], int, List[int]]] = None
    """
    A List of flags, or a bitwise map of flags, modifying this transaction's
    behavior. See `Flags Field
    <https://xrpl.org/transaction-common-fields.html#flags-field>`_ for more details.
    """

    last_ledger_sequence: Optional[int] = None
    """
    The highest ledger index this transaction can appear in. Specifying this
    field places a strict upper limit on how long the transaction can wait
    to be validated or rejected. See `Reliable Transaction Submission
    <https://xrpl.org/reliable-transaction-submission.html>`_ for details.
    """

    memos: Optional[List[Memo]] = None
    """Additional arbitrary information attached to this transaction."""

    signers: Optional[List[Signer]] = None
    """
    Signing data authorizing a multi-signed transaction. Added during
    multi-signing.
    """

    source_tag: Optional[int] = None
    """
    An arbitrary `source tag
    <https://xrpl.org/source-and-destination-tags.html>`_ representing a
    hosted user or specific purpose at the sending account where this
    transaction comes from.
    """

    signing_pub_key: str = ""
    """
    The public key authorizing a single-signed transaction. Automatically
    added during signing.
    """

    ticket_sequence: Optional[int] = None
    """
    The sequence number of the ticket to use in place of a Sequence number. If
    this is provided, sequence must be 0. Cannot be used with account_txn_id.
    """

    txn_signature: Optional[str] = None
    """
    The cryptographic signature from the sender that authorizes this
    transaction. Automatically added during signing.
    """

    network_id: Optional[int] = None
    """The network id of the transaction."""

    delegate: Optional[str] = None
    """The delegate account that is sending the transaction."""

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if self.ticket_sequence is not None and (
            (self.sequence is not None and self.sequence != 0)
            or self.account_txn_id is not None
        ):
            errors[
                "Transaction"
            ] = """If ticket_sequence is provided,
            account_txn_id must be None and sequence must be None or 0"""

        if self.account == self.delegate:
            errors["delegate"] = "Account and delegate addresses cannot be the same"

        return errors

    def to_dict(self: Self) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a Transaction.

        Returns:
            The dictionary representation of a Transaction.
        """
        # we need to override this because transaction_type is using ``field``
        # which will not include the value in the objects __dict__
        prepared_dict = {
            **super().to_dict(),
            "transaction_type": self.transaction_type.value,
        }
        flags = self._flags_to_int()
        if flags is not None:
            prepared_dict["flags"] = flags
        return prepared_dict

    def _iter_to_int(
        self: Self,
        lst: List[int],
    ) -> int:
        """Calculate flag as int."""
        accumulator = 0
        for flag in lst:
            accumulator |= flag
        return accumulator

    def _flags_to_int(self: Self) -> int | None:
        if self.flags is None:
            return None
        if isinstance(self.flags, int):
            return self.flags
        check_false_flag_definition(tx_type=self.transaction_type, tx_flags=self.flags)
        if isinstance(self.flags, dict):
            return self._iter_to_int(
                lst=interface_to_flag_list(
                    tx_type=self.transaction_type,
                    tx_flags=self.flags,
                )
            )

        return self._iter_to_int(lst=self.flags)

    def to_xrpl(self: Self) -> Dict[str, Any]:
        """
        Creates a JSON-like dictionary in the JSON format used by the binary codec
        based on the Transaction object.

        Returns:
            A JSON-like dictionary in the JSON format used by the binary codec.
        """
        return transaction_json_to_binary_codec_form(self.to_dict())

    def blob(self: Self) -> str:
        """
        Creates the canonical binary format of the Transaction object.

        Returns:
            The binary-encoded object, as a hexadecimal string.
        """
        return encode(self.to_xrpl())

    @classmethod
    def from_dict(cls: Type[Self], value: Dict[str, Any]) -> Self:
        """
        Construct a new Transaction from a dictionary of parameters.

        Args:
            value: The value to construct the Transaction from.

        Returns:
            A new Transaction object, constructed using the given parameters.

        Raises:
            XRPLModelException: If the dictionary provided is invalid.
        """
        if cls.__name__ == "Transaction" or cls.__name__ == "PseudoTransaction":
            # using `(Pseudo)Transaction.from_dict` and not a subclass
            if "transaction_type" not in value:
                raise XRPLModelException(
                    "Transaction does not include transaction_type."
                )
            correct_type = cls.get_transaction_type(value["transaction_type"])
            return correct_type.from_dict(value)  # type: ignore
        else:
            if "transaction_type" in value:
                if value["transaction_type"] != cls.__name__:
                    transaction_type = value["transaction_type"]
                    raise XRPLModelException(
                        f"Using wrong constructor: using {cls.__name__} constructor "
                        f"with transaction type {transaction_type}."
                    )
                value = {**value}
                del value["transaction_type"]
            return super().from_dict(value)

    def has_flag(self: Self, flag: int) -> bool:
        """
        Returns whether the transaction has the given flag value set.

        Args:
            flag: The given flag value for which the function will determine whether it
                is set.

        Returns:
            Whether the transaction has the given flag value set.

        Raises:
            XRPLModelException: if `self.flags` is invalid.
        """
        if self.flags is None:
            return False
        if isinstance(self.flags, int):
            return self.flags & flag != 0
        elif isinstance(self.flags, dict):
            return flag in interface_to_flag_list(
                tx_type=self.transaction_type,
                tx_flags=self.flags,
            )
        elif isinstance(self.flags, list):
            return flag in self.flags
        else:
            raise XRPLModelException("self.flags is not an int, dict, or list")

    def is_signed(self: Self) -> bool:
        """
        Checks if a transaction has been signed.

        Returns:
            Whether the transaction has been signed
        """
        if self.signers:
            for signer in self.signers:
                if (
                    signer.signing_pub_key is None or len(signer.signing_pub_key) <= 0
                ) or (signer.txn_signature is None or len(signer.txn_signature) <= 0):
                    return False
            return True
        return (
            self.signing_pub_key is not None and len(self.signing_pub_key) > 0
        ) and (self.txn_signature is not None and len(self.txn_signature) > 0)

    def get_hash(self: Self) -> str:
        """
        Hashes the Transaction object as the ledger does. Only valid for signed
        Transaction objects.

        Returns:
            The hash of the Transaction object.

        Raises:
            XRPLModelException: if the Transaction is unsigned.
        """
        if (
            self.txn_signature is None
            and self.signers is None
            and not self.has_flag(TransactionFlag.TF_INNER_BATCH_TXN)
        ):
            raise XRPLModelException(
                "Cannot get the hash from an unsigned Transaction."
            )
        prefix = hex(_TRANSACTION_HASH_PREFIX)[2:].upper()
        encoded_str = bytes.fromhex(prefix + encode(self.to_xrpl()))
        return sha512(encoded_str).digest().hex().upper()[:64]

    @classmethod
    def get_transaction_type(
        cls: Type[Self], transaction_type: str
    ) -> Type[Transaction]:
        """
        Returns the correct transaction type based on the string name.

        Args:
            transaction_type: The String name of the Transaction object.

        Returns:
            The transaction class with the given name.

        Raises:
            XRPLModelException: If `transaction_type` is not a valid Transaction type.
        """
        import xrpl.models.transactions as transaction_models
        import xrpl.models.transactions.pseudo_transactions as pseudo_transaction_models

        transaction_types: Dict[str, Type[Transaction]] = {
            t.value: getattr(transaction_models, t)
            for t in transaction_models.types.TransactionType
        }
        if transaction_type in transaction_types:
            return transaction_types[transaction_type]

        pseudo_transaction_types: Dict[str, Type[Transaction]] = {
            t.value: getattr(pseudo_transaction_models, t)
            for t in transaction_models.types.PseudoTransactionType
        }
        if transaction_type in pseudo_transaction_types:
            return pseudo_transaction_types[transaction_type]

        raise XRPLModelException(f"{transaction_type} is not a valid Transaction type")

    @staticmethod
    def from_blob(tx_blob: str) -> Transaction:
        """
        Decodes a transaction blob.

        Args:
            tx_blob: the tx blob to decode.

        Returns:
            The formatted transaction.
        """
        return Transaction.from_xrpl(decode(tx_blob))

    @classmethod
    def from_xrpl(cls: Type[Self], value: Union[str, Dict[str, Any]]) -> Self:
        """
        Creates a Transaction object based on a JSON or JSON-string representation of
        data

        In Payment transactions, the DeliverMax field is renamed to the Amount field.

        Args:
            value: The dictionary or JSON string to be instantiated.

        Returns:
            A Transaction object instantiated from the input.

        Raises:
            XRPLModelException: If Payment transactions have different values for
                                amount and deliver_max fields
        """
        processed_value = cls._process_xrpl_json(value)
        # handle the deliver_max alias in Payment transactions
        if (
            "transaction_type" in processed_value
            and processed_value["transaction_type"] == "Payment"
        ) and "deliver_max" in processed_value:
            if (
                "amount" in processed_value
                and processed_value["amount"] != processed_value["deliver_max"]
            ):
                raise XRPLModelException(
                    "Error: amount and deliver_max fields must be equal if both are "
                    + "provided"
                )
            else:
                processed_value["amount"] = processed_value["deliver_max"]

            # deliver_max field is not recognised in the Payment Request format,
            # nor is it supported in the serialization operations.
            del processed_value["deliver_max"]

        return cls.from_dict(processed_value)



"""Model for TicketCreate transaction type."""

from dataclasses import dataclass, field

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class TicketCreate(Transaction):
    """
    A TicketCreate transaction sets aside one or more `sequence numbers
    <https://xrpl.org/basic-data-types.html#account-sequence>`_ as `Tickets
    <https://xrpl.org/tickets.html>`_.
    """

    ticket_count: int = REQUIRED  # type: ignore
    """
    How many Tickets to create. This must be a positive number and cannot cause the
    account to own more than 250 Tickets after executing this transaction.
    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.TICKET_CREATE,
        init=False,
    )



"""The base class for models that involve a nested dictionary e.g. memos."""

from __future__ import annotations

from typing import Any, Dict, Type, Union

from typing_extensions import Self

from xrpl.models.base_model import BaseModel, _key_to_json


def _get_nested_name(cls: Union[NestedModel, Type[NestedModel]]) -> str:
    if isinstance(cls, NestedModel):
        name = cls.__class__.__name__
    else:
        name = cls.__name__
    return _key_to_json(name)


class NestedModel(BaseModel):
    """The base class for models that involve a nested dictionary e.g. memos."""

    @classmethod
    def is_dict_of_model(cls: Type[Self], dictionary: Any) -> bool:  # noqa: ANN401
        """
        Returns True if the input dictionary was derived by the `to_dict`
        method of an instance of this class. In other words, True if this is
        a dictionary representation of an instance of this class.

        NOTE: does not account for model inheritance, IE will only return True
        if dictionary represents an instance of this class, but not if
        dictionary represents an instance of a subclass of this class.

        Args:
            dictionary: The dictionary to check.

        Returns:
            True if dictionary is a dict representation of an instance of this
            class.
        """
        return (
            isinstance(dictionary, dict)
            and _get_nested_name(cls) in dictionary
            and super().is_dict_of_model(dictionary[_get_nested_name(cls)])
        )

    @classmethod
    def from_dict(cls: Type[Self], value: Dict[str, Any]) -> Self:
        """
        Construct a new NestedModel from a dictionary of parameters.

        Args:
            value: The value to construct the NestedModel from.

        Returns:
            A new NestedModel object, constructed using the given parameters.

        Raises:
            XRPLModelException: If the dictionary provided is invalid.
        """
        if _get_nested_name(cls) not in value:
            return super().from_dict(value)
        return super().from_dict(value[_get_nested_name(cls)])

    def to_dict(self: Self) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a NestedModel.

        Returns:
            The dictionary representation of a NestedModel.
        """
        return {_get_nested_name(self): super().to_dict()}



"""Model for AccountSet transaction type."""

from __future__ import annotations  # Requires Python 3.7+

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from typing_extensions import Final, Self

from xrpl.models.transactions.transaction import Transaction, TransactionFlagInterface
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init

_MAX_TRANSFER_RATE: Final[int] = 2000000000
_MIN_TRANSFER_RATE: Final[int] = 1000000000
_SPECIAL_CASE_TRANFER_RATE: Final[int] = 0

_MIN_TICK_SIZE: Final[int] = 3
_MAX_TICK_SIZE: Final[int] = 15
_DISABLE_TICK_SIZE: Final[int] = 0

_MAX_DOMAIN_LENGTH: Final[int] = 256


class AccountSetAsfFlag(int, Enum):
    """
    Enum for AccountSet Flags.

    There are several options which can be either enabled or disabled for an account.
    Account options are represented by different types of flags depending on the
    situation. The AccountSet transaction type has several "AccountSet Flags" (prefixed
    `asf`) that can enable an option when passed as the SetFlag parameter, or disable
    an option when passed as the ClearFlag parameter. This enum represents those
    options.

    `See AccountSet asf Flags <https://xrpl.org/accountset.html#accountset-flags>`_
    """

    ASF_ACCOUNT_TXN_ID = 5
    """
    Track the ID of this account's most recent transaction. Required for
    `AccountTxnID <https://xrpl.org/transaction-common-fields.html#accounttxnid>`_
    """

    ASF_DEFAULT_RIPPLE = 8
    """
    Enable `rippling
    <https://xrpl.org/rippling.html>`_ on this account's trust lines by default.
    """

    ASF_DEPOSIT_AUTH = 9
    """
    Enable `Deposit Authorization
    <https://xrpl.org/depositauth.html>`_ on this account.
    """

    ASF_DISABLE_MASTER = 4
    """
    Disallow use of the master key pair. Can only be enabled if the account has
    configured another way to sign transactions, such as a `Regular Key
    <https://xrpl.org/cryptographic-keys.html>`_ or a `Signer List
    <https://xrpl.org/multi-signing.html>`_.
    """

    ASF_DISALLOW_XRP = 3
    """XRP should not be sent to this account. (Enforced by client applications)"""

    ASF_GLOBAL_FREEZE = 7
    """
    `Freeze
    <https://xrpl.org/freezes.html>`_ all assets issued by this account.
    """

    ASF_NO_FREEZE = 6
    """
    Permanently give up the ability to `freeze individual trust lines or disable
    Global Freeze <https://xrpl.org/freezes.html>`_. This flag can never be disabled
    after being enabled.
    """

    ASF_REQUIRE_AUTH = 2
    """
    Require authorization for users to hold balances issued by this address. Can
    only be enabled if the address has no trust lines connected to it.
    """

    ASF_REQUIRE_DEST = 1
    """Require a destination tag to send transactions to this account."""

    ASF_AUTHORIZED_NFTOKEN_MINTER = 10
    """Allow another account to mint and burn tokens on behalf of this account."""

    ASF_DISABLE_INCOMING_NFTOKEN_OFFER = 12
    """Disallow other accounts from creating NFTokenOffers directed at this account."""

    ASF_DISABLE_INCOMING_CHECK = 13
    """Disallow other accounts from creating Checks directed at this account."""

    ASF_DISABLE_INCOMING_PAYCHAN = 14
    """Disallow other accounts from creating PayChannels directed at this account."""

    ASF_DISABLE_INCOMING_TRUSTLINE = 15
    """Disallow other accounts from creating Trustlines directed at this account."""

    ASF_ALLOW_TRUSTLINE_CLAWBACK = 16
    """Allow trustline clawback feature"""


class AccountSetFlag(int, Enum):
    """
    Enum for AccountSet Transaction Flags.

    Transactions of the AccountSet type support additional values in the Flags field.
    This enum represents those options.

    `See AccountSet tf Flags <https://xrpl.org/accountset.html#accountset-flags>`_
    """

    TF_REQUIRE_DEST_TAG = 0x00010000
    """
    The same as SetFlag: asfRequireDest.
    """

    TF_OPTIONAL_DEST_TAG = 0x00020000
    """
    The same as ClearFlag: asfRequireDest.
    """

    TF_REQUIRE_AUTH = 0x00040000
    """
    The same as SetFlag: asfRequireAuth.
    """

    TF_OPTIONAL_AUTH = 0x00080000
    """
    The same as ClearFlag: asfRequireAuth.
    """

    TF_DISALLOW_XRP = 0x00100000
    """
    The same as SetFlag: asfDisallowXRP.
    """

    TF_ALLOW_XRP = 0x00200000
    """
    The same as ClearFlag: asfDisallowXRP.
    """


class AccountSetFlagInterface(TransactionFlagInterface):
    """
    Transactions of the AccountSet type support additional values in the Flags field.
    This TypedDict represents those options.

    `See AccountSet tf Flags <https://xrpl.org/accountset.html#accountset-flags>`_
    """

    TF_REQUIRE_DEST_TAG: bool
    TF_OPTIONAL_DEST_TAG: bool
    TF_REQUIRE_AUTH: bool
    TF_OPTIONAL_AUTH: bool
    TF_DISALLOW_XRP: bool
    TF_ALLOW_XRP: bool


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class AccountSet(Transaction):
    """
    Represents an `AccountSet transaction <https://xrpl.org/accountset.html>`_,
    which modifies the properties of an account in the XRP Ledger.
    """

    clear_flag: Optional[AccountSetAsfFlag] = None
    """
    Disable a specific `AccountSet Flag
    <https://xrpl.org/accountset.html#accountset-flags>`_
    """

    domain: Optional[str] = None
    """
    Set the DNS domain of the account owner. Must be hex-encoded. You can
    use `xrpl.utils.str_to_hex` to convert a UTF-8 string to hex.
    """

    email_hash: Optional[str] = None
    """
    Set the MD5 Hash to be used for generating an avatar image for this
    account.
    """

    message_key: Optional[str] = None
    """Set a public key for sending encrypted messages to this account."""

    set_flag: Optional[AccountSetAsfFlag] = None
    """
    Enable a specific `AccountSet Flag
    <https://xrpl.org/accountset.html#accountset-flags>`_
    """

    transfer_rate: Optional[int] = None
    """
    Set the transfer fee to use for tokens issued by this account. See
    `TransferRate <https://xrpl.org/accountset.html#transferrate>`_ for
    details.
    """

    tick_size: Optional[int] = None
    """
    Set the tick size to use when trading tokens issued by this account in
    the decentralized exchange. See `Tick Size
    <https://xrpl.org/ticksize.html>`_ for details.
    """

    nftoken_minter: Optional[str] = None
    """
    Sets an alternate account that is allowed to mint NFTokens on this
    account's behalf using NFTokenMint's `Issuer` field. If set, you must
    also set the AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER flag.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.ACCOUNT_SET,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "tick_size": self._get_tick_size_error(),
                "transfer_rate": self._get_transfer_rate_error(),
                "domain": self._get_domain_error(),
                "clear_flag": self._get_clear_flag_error(),
                "nftoken_minter": self._get_nftoken_minter_error(),
            }.items()
            if value is not None
        }

    def _get_tick_size_error(self: Self) -> Optional[str]:
        if self.tick_size is None:
            return None
        if self.tick_size > _MAX_TICK_SIZE:
            return f"`tick_size` is above {_MAX_TICK_SIZE}."
        if self.tick_size < _MIN_TICK_SIZE and self.tick_size != _DISABLE_TICK_SIZE:
            return f"`tick_size` is below {_MIN_TICK_SIZE}."
        return None

    def _get_transfer_rate_error(self: Self) -> Optional[str]:
        if self.transfer_rate is None:
            return None
        if self.transfer_rate > _MAX_TRANSFER_RATE:
            return f"`transfer_rate` is above {_MAX_TRANSFER_RATE}."
        if (
            self.transfer_rate < _MIN_TRANSFER_RATE
            and self.transfer_rate != _SPECIAL_CASE_TRANFER_RATE
        ):
            return f"`transfer_rate` is below {_MIN_TRANSFER_RATE}."
        return None

    def _get_domain_error(self: Self) -> Optional[str]:
        if self.domain is not None and self.domain.lower() != self.domain:
            return f"Domain {self.domain} is not lowercase"
        if self.domain is not None and len(self.domain) > _MAX_DOMAIN_LENGTH:
            return f"Must not be longer than {_MAX_DOMAIN_LENGTH} characters"
        return None

    def _get_clear_flag_error(self: Self) -> Optional[str]:
        if self.clear_flag is not None and self.clear_flag == self.set_flag:
            return "Must not be equal to the set_flag"
        return None

    def _get_nftoken_minter_error(self: Self) -> Optional[str]:
        if (
            self.set_flag != AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER
            and self.nftoken_minter is not None
        ):
            return (
                "Will not set the minter unless "
                "AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER is set"
            )
        if (
            self.set_flag == AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER
            and self.nftoken_minter is None
        ):
            return (
                "Must be present if AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER "
                "is set"
            )
        if (
            self.clear_flag == AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER
            and self.nftoken_minter is not None
        ):
            return (
                "Must not be present if "
                "AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER is unset "
                "using clear_flag"
            )
        return None



"""High-level transaction methods with XRPL transactions."""

import math
from typing import Any, Dict, List, Optional, Union, cast

from typing_extensions import Final, TypeVar

from xrpl.asyncio.account import get_next_valid_seq_number
from xrpl.asyncio.clients import Client, XRPLRequestFailureException
from xrpl.asyncio.clients.client import get_network_id_and_build_version
from xrpl.asyncio.ledger import get_fee, get_latest_validated_ledger_sequence
from xrpl.constants import XRPLException
from xrpl.core.addresscodec import is_valid_xaddress, xaddress_to_classic_address
from xrpl.core.binarycodec import encode, encode_for_multisigning, encode_for_signing
from xrpl.core.keypairs.main import sign as keypairs_sign
from xrpl.models import (
    Batch,
    EscrowFinish,
    Response,
    ServerState,
    Simulate,
    SubmitOnly,
    Transaction,
    TransactionFlag,
)
from xrpl.models.transactions.transaction import (
    transaction_json_to_binary_codec_form as model_transaction_to_binary_codec,
)
from xrpl.models.transactions.types.transaction_type import TransactionType
from xrpl.utils import drops_to_xrp, xrp_to_drops
from xrpl.wallet import Wallet

_LEDGER_OFFSET: Final[int] = 20
# Sidechains are expected to have network IDs above this.
# Networks with ID above this restricted number are expected to specify an
# accurate NetworkID field in every transaction to that chain to prevent replay attacks.
# Mainnet and testnet are exceptions.
# More context: https://github.com/XRPLF/rippled/pull/4370
_RESTRICTED_NETWORKS = 1024
_REQUIRED_NETWORKID_VERSION = "1.11.0"

T = TypeVar("T", bound=Transaction, default=Transaction)


async def sign_and_submit(
    transaction: Transaction,
    client: Client,
    wallet: Wallet,
    autofill: bool = True,
    check_fee: bool = True,
) -> Response:
    """
    Signs a transaction (locally, without trusting external rippled nodes) and submits
    it to the XRPL.

    Args:
        transaction: the transaction to be signed and submitted.
        client: the network client with which to submit the transaction.
        wallet: the wallet with which to sign the transaction.
        autofill: whether to autofill the relevant fields. Defaults to True.
        check_fee: whether to check if the fee is higher than the expected transaction
            type fee. Defaults to True.

    Returns:
        The response from the ledger.
    """
    if autofill:
        transaction = await autofill_and_sign(transaction, client, wallet, check_fee)
    else:
        if check_fee:
            await _check_fee(transaction, client)
        transaction = sign(transaction, wallet)
    return await submit(transaction, client)


# Even though this is synchronous - this is here because it used to be async in
# xrpl-py 1.0, and we decided it wasn't worth breaking people's imports to move
# It to a central location as part of the xrpl-py 2.0 changes. It is aliased in
# The synchronous half of the library as well.
def sign(
    transaction: T,
    wallet: Wallet,
    multisign: bool = False,
) -> T:
    """
    Signs a transaction locally, without trusting external rippled nodes.

    Args:
        transaction: the transaction to be signed.
        wallet: the wallet with which to sign the transaction.
        multisign: whether to sign the transaction for a multisignature transaction.

    Returns:
        The signed transaction blob.
    """
    transaction_json = _prepare_transaction(transaction)
    if multisign:
        signature = keypairs_sign(
            bytes.fromhex(
                encode_for_multisigning(
                    transaction_json,
                    wallet.address,
                )
            ),
            wallet.private_key,
        )
        transaction_json["Signers"] = [
            {
                "Signer": {
                    "Account": wallet.address,
                    "TxnSignature": signature,
                    "SigningPubKey": wallet.public_key,
                }
            }
        ]
        return cast(T, Transaction.from_xrpl(transaction_json))

    transaction_json["SigningPubKey"] = wallet.public_key
    serialized_for_signing = encode_for_signing(transaction_json)
    serialized_bytes = bytes.fromhex(serialized_for_signing)
    signature = keypairs_sign(serialized_bytes, wallet.private_key)
    transaction_json["TxnSignature"] = signature
    return cast(T, Transaction.from_xrpl(transaction_json))


async def autofill_and_sign(
    transaction: T,
    client: Client,
    wallet: Wallet,
    check_fee: bool = True,
) -> T:
    """
    Autofills relevant fields. Then, signs a transaction locally, without trusting
    external rippled nodes.

    Args:
        transaction: the transaction to be signed.
        wallet: the wallet with which to sign the transaction.
        client: a network client.
        check_fee: whether to check if the fee is higher than the expected transaction
            type fee. Defaults to True.

    Returns:
        The signed transaction.
    """
    # We do the transaction fee check here as we have the Client available.
    # The fee check will be done if transaction.fee exists. Otherwise the fee
    # will be auto-filled in autofill()
    if check_fee:
        await _check_fee(transaction, client)

    return sign(await autofill(transaction, client), wallet, multisign=False)


async def submit(
    transaction: Transaction,
    client: Client,
    *,
    fail_hard: bool = False,
) -> Response:
    """
    Submits a transaction to the ledger.

    Args:
        transaction: The Transaction to be submitted.
        client: The network client with which to submit the transaction.
        fail_hard: An optional boolean. If True, and the transaction fails for
            the initial server, do not retry or relay the transaction to other
            servers. Defaults to False.

    Returns:
        The response from the ledger.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """
    transaction_blob = encode(transaction.to_xrpl())
    response = await client._request_impl(
        SubmitOnly(tx_blob=transaction_blob, fail_hard=fail_hard)
    )
    if response.is_successful():
        return response

    raise XRPLRequestFailureException(response.result)


async def simulate(
    transaction: Transaction,
    client: Client,
    *,
    binary: bool = False,
) -> Response:
    """
    Simulates a transaction without actually submitting it to the network.

    Args:
        transaction: The transaction to simulate.
        client: The network client with which to submit the transaction.
        binary: Whether the return data should be encoded in the XRPL's binary format.
            Defaults to False.

    Raises:
        XRPLRequestFailureException: If the transaction fails in the simulated scenario.

    Returns:
        The response from the ledger.
    """
    # autofill the network ID
    transaction_json = transaction.to_dict()
    await get_network_id_and_build_version(client)
    if "network_id" not in transaction_json and _tx_needs_networkID(client):
        transaction_json["network_id"] = client.network_id
    final_tx = Transaction.from_dict(transaction_json)

    # send the `simulate` request
    response = await client._request_impl(Simulate(transaction=final_tx, binary=binary))
    if response.is_successful():
        return response

    raise XRPLRequestFailureException(response.result)


def _prepare_transaction(transaction: Transaction) -> Dict[str, Any]:
    """
    Prepares a Transaction by converting it to a JSON-like dictionary, converting the
    field names to CamelCase. If a Client is provided, then it also autofills any
    relevant fields.

    Args:
        transaction: the Transaction to be prepared.

    Returns:
        A JSON-like dictionary that is ready to be signed.

    Raises:
        XRPLException: if both LastLedgerSequence and `ledger_offset` are provided, or
            if an address tag is provided that does not match the X-Address tag, or if
            attempting to directly sign a Batch inner transaction.
    """
    if transaction.has_flag(TransactionFlag.TF_INNER_BATCH_TXN):
        raise XRPLException("Cannot directly sign a batch inner transaction.")

    transaction_json = transaction.to_xrpl()

    _validate_account_xaddress(transaction_json, "Account", "SourceTag")
    if "Destination" in transaction_json:
        _validate_account_xaddress(transaction_json, "Destination", "DestinationTag")

    # DepositPreauth
    _convert_to_classic_address(transaction_json, "Authorize")
    _convert_to_classic_address(transaction_json, "Unauthorize")
    # EscrowCancel, EscrowFinish
    _convert_to_classic_address(transaction_json, "Owner")
    # SetRegularKey
    _convert_to_classic_address(transaction_json, "RegularKey")

    return transaction_json


async def autofill(
    transaction: T, client: Client, signers_count: Optional[int] = None
) -> T:
    """
    Autofills fields in a transaction. This will set all autofill-able fields according
    to the current state of the server this Client is connected to. For Batch
    transactions, it will also handle autofilling inner transactions. It also converts
    all X-Addresses to classic addresses.

    Args:
        transaction: the transaction to be signed.
        client: a network client.
        signers_count: the expected number of signers for this transaction.
            Only used for multisigned transactions.

    Raises:
        XRPLException: If a field is pre-filled out incorrectly.

    Returns:
        The autofilled transaction.
    """
    transaction_json = transaction.to_dict()
    await get_network_id_and_build_version(client)
    if "network_id" not in transaction_json and _tx_needs_networkID(client):
        transaction_json["network_id"] = client.network_id
    if "sequence" not in transaction_json:
        if "ticket_sequence" in transaction_json:
            sequence = 0
        else:
            sequence = await get_next_valid_seq_number(
                transaction_json["account"], client
            )
        transaction_json["sequence"] = sequence
    if "fee" not in transaction_json:
        transaction_json["fee"] = await _calculate_fee_per_transaction_type(
            transaction, client, signers_count
        )
    if "last_ledger_sequence" not in transaction_json:
        ledger_sequence = await get_latest_validated_ledger_sequence(client)
        transaction_json["last_ledger_sequence"] = ledger_sequence + _LEDGER_OFFSET
    if transaction.transaction_type == TransactionType.BATCH:
        inner_txs = await _autofill_batch(client, transaction_json)
        transaction_json["raw_transactions"] = inner_txs
    return cast(T, Transaction.from_dict(transaction_json))


def _tx_needs_networkID(client: Client) -> bool:
    """
    Determines whether the transactions required network ID to be valid.
    Transaction needs networkID if later than restricted ID and either
        the network is hooks testnet or build version is >= 1.11.0.
    More context: https://github.com/XRPLF/rippled/pull/4370

    Args:
        client (Client): The network client to use to send the request.

    Returns:
        bool: whether the transactions required network ID to be valid
    """
    if client.network_id is not None and client.network_id > _RESTRICTED_NETWORKS:
        if client.build_version and _is_not_later_rippled_version(
            _REQUIRED_NETWORKID_VERSION, client.build_version
        ):
            return True
    return False


def _is_not_later_rippled_version(source: str, target: str) -> bool:
    """
    Determines whether the source version is not a later release than the
        target version.

    Args:
        source: the source rippled version.
        target: the target rippled version.

    Returns:
        bool: true if source is earlier, false otherwise.
    """
    if source == target:
        return True
    source_decomp = source.split(".")
    target_decomp = target.split(".")
    source_major, source_minor = int(source_decomp[0]), int(source_decomp[1])
    target_major, target_minor = int(target_decomp[0]), int(target_decomp[1])

    # Compare major version
    if source_major != target_major:
        return source_major < target_major

    # Compare minor version
    if source_minor != target_minor:
        return source_minor < target_minor

    source_patch = source_decomp[2].split("-")
    target_patch = target_decomp[2].split("-")
    source_patch_version = int(source_patch[0])
    target_patch_version = int(target_patch[0])

    # Compare patch version
    if source_patch_version != target_patch_version:
        return source_patch_version < target_patch_version

    # Compare release version
    if len(source_patch) != len(target_patch):
        return len(source_patch) > len(target_patch)

    if len(source_patch) == 2:
        # Compare release types
        if not source_patch[1][0].startswith(target_patch[1][0]):
            return source_patch[1] < target_patch[1]
        # Compare beta versions
        if source_patch[1].startswith("b"):
            return int(source_patch[1][1:]) < int(target_patch[1][1:])
        # Compare rc versions
        return int(source_patch[1][2:]) < int(target_patch[1][2:])
    return False


def _validate_account_xaddress(
    json: Dict[str, Any], account_field: str, tag_field: str
) -> None:
    """
    Mutates JSON-like dictionary so the X-Address in the account field is the classic
    address, and the tag is in the tag field.

    Args:
        json: JSON-like dictionary with transaction data or similar
        account_field: the field of `json` that may contain an X-Address
        tag_field: the field of `json` that may contain a source or destination tag

    Raises:
        XRPLException: if both an X-Address containing a tag and a tag field are
            provided and they do not match.
    """
    if is_valid_xaddress(json[account_field]):
        account, tag, _ = xaddress_to_classic_address(json[account_field])
        json[account_field] = account
        if tag_field in json and json[tag_field] != tag:
            raise XRPLException(f"{tag_field} value does not match X-Address tag")
        json[tag_field] = tag


def _convert_to_classic_address(json: Dict[str, Any], field: str) -> None:
    """
    Mutates JSON-like dictionary to convert the given field from an X-Address (if
    applicable) to a classic address.

    Args:
        json: JSON-like dictionary with transaction data or similar
        field: the field in `json` that may contain an X-Address
    """
    if field in json and is_valid_xaddress(json[field]):
        json[field] = xaddress_to_classic_address(json[field])[0]


def transaction_json_to_binary_codec_form(dictionary: Dict[str, Any]) -> Dict[str, Any]:
    """
    Returns a new dictionary in which the keys have been formatted as CamelCase and
    standardized to be serialized by the binary codec.

    Args:
        dictionary: The dictionary to be reformatted.

    Returns:
        A new dictionary object that has been reformatted.
    """
    return model_transaction_to_binary_codec(dictionary)


async def _check_fee(
    transaction: Transaction,
    client: Client,
    signers_count: Optional[int] = None,
) -> None:
    """
    Checks if the Transaction fee is higher than the expected Transaction type fee.

    Args:
        transaction: The transaction to check.
        client: Client instance to use to look up network load
        signers_count: the expected number of signers for this transaction.
            Only used for multisigned transactions.

    Raises:
        XRPLException: if the transaction fee is higher than the expected fee.
    """
    expected_fee = max(
        int(xrp_to_drops(0.1)),  # a fee that is obviously too high
        int(
            await _calculate_fee_per_transaction_type(
                transaction, client, signers_count
            )
        ),
    )

    if transaction.fee and int(transaction.fee) > expected_fee:
        raise XRPLException(
            f"Fee value: {str(drops_to_xrp(transaction.fee))} XRP is likely entered "
            "incorrectly, since it is much larger than the typical XRP transaction "
            "cost. If this is intentional, use `check_fee=False`."
        )


async def _calculate_fee_per_transaction_type(
    transaction: Transaction,
    client: Client,
    signers_count: Optional[int] = None,
) -> str:
    """
    Calculate the total fee in drops for a transaction based on:
    - the network fee
    - the transaction condition

    https://xrpl.org/transaction-cost.html#special-transaction-costs

    Args:
        transaction: the Transaction to be submitted.
        client: the network client with which to submit the transaction.
        signers_count: the expected number of signers for this transaction.
            Only used for multisigned transactions and multi-account/multi-signed Batch
            transactions.

    Returns:
        The expected Transaction fee in drops
    """
    # Reference Transaction (Most transactions)

    net_fee = int(
        await get_fee(client)
    )  # Latest data is found in FeeSettings ledger-object's BaseFee field.

    base_fee = net_fee

    # EscrowFinish Transaction with Fulfillment
    # https://xrpl.org/escrowfinish.html#escrowfinish-fields
    if transaction.transaction_type == TransactionType.ESCROW_FINISH:
        escrow_finish = cast(EscrowFinish, transaction)
        if escrow_finish.fulfillment is not None:
            fulfillment_bytes = escrow_finish.fulfillment.encode("ascii")
            # BaseFee × (33 + (Fulfillment size in bytes / 16))
            base_fee = math.ceil(net_fee * (33 + (len(fulfillment_bytes) / 16)))

    # AccountDelete Transaction
    elif transaction.transaction_type in (
        TransactionType.ACCOUNT_DELETE,
        TransactionType.AMM_CREATE,
    ):
        base_fee = await _fetch_owner_reserve_fee(client)

    elif transaction.transaction_type == TransactionType.BATCH:
        batch = cast(Batch, transaction)
        base_fee = base_fee * 2 + sum(
            [
                int(await _calculate_fee_per_transaction_type(raw_txn, client))
                for raw_txn in batch.raw_transactions
            ]
        )

    # Multi-signed/Multi-Account Batch Transactions
    # BaseFee × (1 + Number of Signatures Provided)
    if signers_count is not None and signers_count > 0:
        base_fee += net_fee * signers_count
    # Round Up base_fee and return it as a String
    return str(math.ceil(base_fee))


async def _fetch_owner_reserve_fee(client: Client) -> int:
    server_state = await client._request_impl(ServerState())
    fee = server_state.result["state"]["validated_ledger"]["reserve_inc"]
    return int(fee)


async def _autofill_batch(
    client: Client, transaction_dict: Dict[str, Any]
) -> List[Dict[str, Any]]:
    transaction = Batch.from_dict(transaction_dict)
    assert transaction.sequence is not None
    account_sequences: Dict[str, int] = {transaction.account: transaction.sequence + 1}
    inner_txs: List[Dict[str, Any]] = []

    for raw_txn in transaction.raw_transactions:
        raw_txn_dict = raw_txn.to_dict()

        if raw_txn.transaction_type == TransactionType.BATCH:
            raise XRPLException(
                "Cannot have a Batch transaction inside a Batch transaction."
            )
        if raw_txn.sequence is None and raw_txn.ticket_sequence is None:
            # autofill sequence
            if raw_txn.account in account_sequences:
                raw_txn_dict["sequence"] = account_sequences[raw_txn.account]
                account_sequences[raw_txn.account] += 1
            else:
                sequence = await get_next_valid_seq_number(raw_txn.account, client)
                account_sequences[raw_txn.account] = sequence + 1
                raw_txn_dict["sequence"] = sequence

        if raw_txn.is_signed():
            raise XRPLException("Inner Batch transactions must not be signed.")

        # validate fields that are supposed to be empty/zeroed
        def _validate_field(field_name: str, expected_value: Union[str, None]) -> None:
            if field_name not in raw_txn_dict:
                raw_txn_dict[field_name] = expected_value
            elif raw_txn_dict[field_name] != expected_value:
                raise XRPLException(
                    f"Must have a `{field_name}` of {repr(expected_value)} in an "
                    "inner Batch transaction."
                )

        _validate_field("fee", "0")
        _validate_field("signing_pub_key", "")
        _validate_field("txn_signature", None)

        if raw_txn.txn_signature is not None:
            raise XRPLException(
                "Must not have a `txn_signature` field in an inner Batch transaction."
            )
        if raw_txn.signers is not None:
            raise XRPLException(
                "Must not have a `signers` field in an inner Batch transaction."
            )
        if raw_txn.network_id is None and _tx_needs_networkID(client):
            raw_txn_dict["network_id"] = client.network_id
        if raw_txn.last_ledger_sequence is not None:
            raise XRPLException(
                "Must not have a `last_ledger_sequence` field in an inner Batch "
                "transaction."
            )

        inner_txs.append(raw_txn_dict)

    return inner_txs



"""Base class for serializing and deserializing unsigned integers.
See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
"""

from __future__ import annotations

from typing import Union

from typing_extensions import Self

from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.serialized_type import SerializedType


class UInt(SerializedType):
    """Base class for serializing and deserializing unsigned integers.
    See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
    """

    def __init__(self: Self, buffer: bytes) -> None:
        """Construct a new UInt type from a ``bytes`` value."""
        self.buffer = buffer

    @property
    def value(self: Self) -> int:
        """
        Get the value of the UInt represented by `self.buffer`.

        Returns:
            The int value of the UInt.
        """
        return int.from_bytes(self.buffer, byteorder="big", signed=False)

    def __eq__(self: Self, other: object) -> bool:
        """Determine whether two UInt objects are equal."""
        if isinstance(other, int):
            return self.value == other
        if isinstance(other, UInt):
            return self.value == other.value
        raise XRPLBinaryCodecException(f"Cannot compare UInt and {type(other)}")

    def __ne__(self: Self, other: object) -> bool:
        """Determine whether two UInt objects are unequal."""
        if isinstance(other, int):
            return self.value != other
        if isinstance(other, UInt):
            return self.value != other.value
        raise XRPLBinaryCodecException(f"Cannot compare UInt and {type(other)}")

    def __lt__(self: Self, other: object) -> bool:
        """Determine whether one UInt object is less than another."""
        if isinstance(other, int):
            return self.value < other
        if isinstance(other, UInt):
            return self.value < other.value
        raise XRPLBinaryCodecException(f"Cannot compare UInt and {type(other)}")

    def __le__(self: Self, other: object) -> bool:
        """Determine whether one UInt object is less than or equal to another."""
        if isinstance(other, int):
            return self.value <= other
        if isinstance(other, UInt):
            return self.value <= other.value
        raise XRPLBinaryCodecException(f"Cannot compare UInt and {type(other)}")

    def __gt__(self: Self, other: object) -> bool:
        """Determine whether one UInt object is greater than another."""
        if isinstance(other, int):
            return self.value > other
        if isinstance(other, UInt):
            return self.value > other.value
        raise XRPLBinaryCodecException(f"Cannot compare UInt and {type(other)}")

    def __ge__(self: Self, other: object) -> bool:
        """Determine whether one UInt object is greater than or equal to another."""
        if isinstance(other, int):
            return self.value >= other
        if isinstance(other, UInt):
            return self.value >= other.value
        raise XRPLBinaryCodecException(f"Cannot compare UInt and {type(other)}")

    def to_json(self: Self) -> Union[str, int]:
        """
        Convert a UInt object to JSON.

        Returns:
            The JSON representation of the UInt object.
        """
        if isinstance(self.value, int):
            return self.value
        return str(self.value)



"""Model for CheckCreate transaction type."""

from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class CheckCreate(Transaction):
    """
    Represents a `CheckCreate <https://xrpl.org/checkcreate.html>`_ transaction,
    which creates a Check object. A Check object is a deferred payment
    that can be cashed by its intended destination. The sender of this
    transaction is the sender of the Check.
    """

    destination: str = REQUIRED  # type: ignore
    """
    The address of the `account
    <https://xrpl.org/accounts.html>`_ that can cash the Check. This field is
    required.

    :meta hide-value:
    """

    send_max: Amount = REQUIRED  # type: ignore
    """
    Maximum amount of source token the Check is allowed to debit the
    sender, including transfer fees on non-XRP tokens. The Check can only
    credit the destination with the same token (from the same issuer, for
    non-XRP tokens). This field is required.

    :meta hide-value:
    """

    destination_tag: Optional[int] = None
    """
    An arbitrary `destination tag
    <https://xrpl.org/source-and-destination-tags.html>`_ that
    identifies the reason for the Check, or a hosted recipient to pay.
    """

    expiration: Optional[int] = None
    """
    Time after which the Check is no longer valid, in seconds since the
    Ripple Epoch.
    """

    invoice_id: Optional[str] = None
    """
    Arbitrary 256-bit hash representing a specific reason or identifier for
    this Check.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.CHECK_CREATE,
        init=False,
    )



"""Model for Batch transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Type

from typing_extensions import Self

from xrpl.models.exceptions import XRPLModelException
from xrpl.models.nested_model import NestedModel
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import (
    Signer,
    Transaction,
    TransactionFlag,
    TransactionFlagInterface,
)
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


class BatchFlag(int, Enum):
    """
    Transactions of the Batch type support additional values in the Flags field.
    This enum represents those options.
    """

    TF_ALL_OR_NOTHING = 0x00010000

    TF_ONLY_ONE = 0x00020000

    TF_UNTIL_FAILURE = 0x00040000

    TF_INDEPENDENT = 0x00080000


class BatchFlagInterface(TransactionFlagInterface):
    """
    Transactions of the Batch type support additional values in the Flags field.
    This TypedDict represents those options.
    """

    TF_ALL_OR_NOTHING: bool
    TF_ONLY_ONE: bool
    TF_UNTIL_FAILURE: bool
    TF_INDEPENDENT: bool


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class BatchSigner(NestedModel):
    """Represents a Batch signer."""

    account: str = REQUIRED  # type: ignore

    signing_pub_key: Optional[str] = None

    txn_signature: Optional[str] = None

    signers: Optional[List[Signer]] = None


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Batch(Transaction):
    """Represents a Batch transaction."""

    raw_transactions: List[Transaction] = REQUIRED  # type: ignore
    batch_signers: Optional[List[BatchSigner]] = None

    transaction_type: TransactionType = field(
        default=TransactionType.BATCH,
        init=False,
    )

    def __post_init__(self: Self) -> None:
        """Called by dataclasses immediately after __init__."""
        new_raw_transactions = []
        for tx in self.raw_transactions:
            # Ensure that every inner transaction has the tfInnerBatchTxn flag set.
            # If it does not, set it.
            if tx.has_flag(TransactionFlag.TF_INNER_BATCH_TXN):
                new_raw_transactions.append(tx)
            else:
                tx_dict = tx.to_dict()
                if "flags" not in tx_dict:
                    tx_dict["flags"] = TransactionFlag.TF_INNER_BATCH_TXN
                elif isinstance(tx_dict["flags"], int):
                    tx_dict["flags"] |= TransactionFlag.TF_INNER_BATCH_TXN
                elif isinstance(tx_dict["flags"], dict):
                    tx_dict["flags"]["TF_INNER_BATCH_TXN"] = True
                elif isinstance(self.flags, list):
                    tx_dict["flags"].append(TransactionFlag.TF_INNER_BATCH_TXN)
                elif tx_dict["flags"] is None:
                    tx_dict["flags"] = TransactionFlag.TF_INNER_BATCH_TXN
                else:
                    raise XRPLModelException(
                        "Invalid `flags` type for raw transaction in Batch."
                    )
                new_raw_transactions.append(Transaction.from_dict(tx_dict))
        # This is done before dataclass locks in the frozen fields
        # This way of editing is a bit hacky, but it's the recommended method for
        # updating frozen fields in dataclasses
        # https://docs.python.org/3/library/dataclasses.html#frozen-instances
        object.__setattr__(self, "raw_transactions", new_raw_transactions)
        super().__post_init__()

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        for i in range(len(self.raw_transactions)):
            tx = self.raw_transactions[i]
            if not tx.has_flag(TransactionFlag.TF_INNER_BATCH_TXN):
                errors[f"raw_transactions[{i}]"] = (
                    "RawTransaction must have tfInnerBatchTxn flag set."
                )
        if len(self.raw_transactions) < 2:
            errors["raw_transactions"] = "Batch must contain at least 2 transactions."
        return errors

    @classmethod
    def from_dict(cls: Type[Self], value: Dict[str, Any]) -> Self:
        """
        Construct a new Batch from a dictionary of parameters.

        Args:
            value: The value to construct the Batch from.

        Returns:
            A new Batch object, constructed using the given parameters.

        Raises:
            XRPLModelException: If the dictionary provided is invalid.
        """
        new_value = {**value}
        new_value["raw_transactions"] = [
            (
                tx["raw_transaction"]
                if isinstance(tx, dict) and "raw_transaction" in tx
                else tx
            )
            for tx in value["raw_transactions"]
        ]
        return super(Transaction, cls).from_dict(new_value)

    def to_dict(self: Self) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a Batch.

        Returns:
            The dictionary representation of a Batch.
        """
        tx_dict = super().to_dict()
        tx_dict["raw_transactions"] = [
            {"raw_transaction": tx} for tx in tx_dict["raw_transactions"]
        ]
        return tx_dict



"""Convenience utilities for the XRP Ledger"""

from xrpl.utils.get_nftoken_id import get_nftoken_id
from xrpl.utils.get_xchain_claim_id import get_xchain_claim_id
from xrpl.utils.parse_nftoken_id import parse_nftoken_id
from xrpl.utils.str_conversions import hex_to_str, str_to_hex
from xrpl.utils.time_conversions import (
    XRPLTimeRangeException,
    datetime_to_ripple_time,
    posix_to_ripple_time,
    ripple_time_to_datetime,
    ripple_time_to_posix,
)
from xrpl.utils.txn_parser import (
    get_balance_changes,
    get_final_balances,
    get_order_book_changes,
)
from xrpl.utils.xrp_conversions import XRPRangeException, drops_to_xrp, xrp_to_drops

__all__ = [
    "str_to_hex",
    "hex_to_str",
    "xrp_to_drops",
    "drops_to_xrp",
    "ripple_time_to_datetime",
    "datetime_to_ripple_time",
    "ripple_time_to_posix",
    "posix_to_ripple_time",
    "XRPRangeException",
    "XRPLTimeRangeException",
    "get_balance_changes",
    "get_final_balances",
    "get_order_book_changes",
    "get_nftoken_id",
    "parse_nftoken_id",
    "get_xchain_claim_id",
]



"""
Specifies an amount in an issued currency.

See https://xrpl.org/currency-formats.html#issued-currency-amounts.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Union

from typing_extensions import Self

from xrpl.models.currencies import IssuedCurrency
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class IssuedCurrencyAmount(IssuedCurrency):
    """
    Specifies an amount in an issued currency.

    See https://xrpl.org/currency-formats.html#issued-currency-amounts.
    """

    value: Union[str, int, float] = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    def to_currency(self: Self) -> IssuedCurrency:
        """
        Build an IssuedCurrency from this IssuedCurrencyAmount.

        Returns:
            The IssuedCurrency for this IssuedCurrencyAmount.
        """
        return IssuedCurrency(issuer=self.issuer, currency=self.currency)

    def to_dict(self: Self) -> Dict[str, str]:
        """
        Returns the dictionary representation of an IssuedCurrencyAmount.

        Returns:
            The dictionary representation of an IssuedCurrencyAmount.
        """
        return {**super().to_dict(), "value": str(self.value)}



"""Utils to get an XChainClaimID from metadata."""

from xrpl.models.transactions.metadata import TransactionMetadata, isCreatedNode


def get_xchain_claim_id(meta: TransactionMetadata) -> str:
    """
    Gets the XChainClaimID from a recently-submitted XChainCreateClaimID transaction.

    Args:
        meta: Metadata from the response to submitting an XChainCreateClaimID
            transaction.

    Returns:
        The newly created XChainClaimID.

    Raises:
        TypeError: if given something other than metadata (e.g. the full
                    transaction response).
    """
    if meta is None or meta.get("AffectedNodes") is None:
        raise TypeError(
            f"""Unable to parse the parameter given to get_xchain_claim_id.
            'meta' must be the metadata from an XChainCreateClaimID transaction.
            Received {meta} instead."""
        )

    affected_nodes = [
        node
        for node in meta["AffectedNodes"]
        if isCreatedNode(node)
        and node["CreatedNode"]["LedgerEntryType"] == "XChainOwnedClaimID"
    ]

    if len(affected_nodes) == 0:
        raise TypeError("No XChainOwnedClaimID created.")

    if len(affected_nodes) > 1:
        # Sanity check - should never happen
        raise TypeError(
            "Multiple XChainOwnedClaimIDs were somehow created. Please report this "
            "error."
        )

    # Get the NFTokenID which wasn't there before this transaction completed.
    return affected_nodes[0]["CreatedNode"]["NewFields"]["XChainClaimID"]



"""Model for NFTokenMint transaction type and related flags."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from typing_extensions import Final, Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction, TransactionFlagInterface
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init

_MAX_URI_LENGTH: Final[int] = 512
_MAX_TRANSFER_FEE: Final[int] = 50000


class NFTokenMintFlag(int, Enum):
    """Transaction Flags for an NFTokenMint Transaction."""

    TF_BURNABLE = 0x00000001
    """
    If set, indicates that the minted token may be burned by the issuer even
    if the issuer does not currently hold the token. The current holder of
    the token may always burn it.
    """

    TF_ONLY_XRP = 0x00000002
    """
    If set, indicates that the token may only be offered or sold for XRP.
    """

    TF_TRUSTLINE = 0x00000004
    """
    If set, indicates that the issuer wants a trustline to be automatically
    created.
    """

    TF_TRANSFERABLE = 0x00000008
    """
    If set, indicates that this NFT can be transferred. This flag has no
    effect if the token is being transferred from the issuer or to the
    issuer.
    """

    TF_MUTABLE = 0x00000010
    """
    If set, indicates that this NFT's URI can be modified.
    """


class NFTokenMintFlagInterface(TransactionFlagInterface):
    """Transaction Flags for an NFTokenMint Transaction."""

    TF_BURNABLE: bool
    TF_ONLY_XRP: bool
    TF_TRUSTLINE: bool
    TF_TRANSFERABLE: bool
    TF_MUTABLE: bool


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class NFTokenMint(Transaction):
    """
    The NFTokenMint transaction creates an NFToken object and adds it to the
    relevant NFTokenPage object of the minter. If the transaction is
    successful, the newly minted token will be owned by the minter account
    specified by the transaction.
    """

    nftoken_taxon: int = REQUIRED  # type: ignore
    """
    Indicates the taxon associated with this token. The taxon is generally a
    value chosen by the minter of the token and a given taxon may be used for
    multiple tokens. The implementation reserves taxon identifiers greater
    than or equal to 2147483648 (0x80000000). If you have no use for this
    field, set it to 0.

    :meta hide-value:
    """

    issuer: Optional[str] = None
    """
    Indicates the account that should be the issuer of this token. This value
    is optional and should only be specified if the account executing the
    transaction is not the `Issuer` of the `NFToken` object. If it is
    present, the `MintAccount` field in the `AccountRoot` of the `Issuer`
    field must match the `Account`, otherwise the transaction will fail.
    """

    transfer_fee: Optional[int] = None
    """
    Specifies the fee charged by the issuer for secondary sales of the Token,
    if such sales are allowed. Valid values for this field are between 0 and
    50000 inclusive, allowing transfer rates between 0.000% and 50.000% in
    increments of 0.001%. This field must NOT be present if the
    `tfTransferable` flag is not set.
    """

    uri: Optional[str] = None
    """
    URI that points to the data and/or metadata associated with the NFT.
    This field need not be an HTTP or HTTPS URL; it could be an IPFS URI, a
    magnet link, immediate data encoded as an RFC2379 "data" URL, or even an
    opaque issuer-specific encoding. The URI is not checked for validity.

    This field must be hex-encoded. You can use `xrpl.utils.str_to_hex` to
    convert a UTF-8 string to hex.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.NFTOKEN_MINT,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "issuer": self._get_issuer_error(),
                "transfer_fee": self._get_transfer_fee_error(),
                "uri": self._get_uri_error(),
            }.items()
            if value is not None
        }

    def _get_issuer_error(self: Self) -> Optional[str]:
        if self.issuer == self.account:
            return "Must not be the same as the account"
        return None

    def _get_transfer_fee_error(self: Self) -> Optional[str]:
        if self.transfer_fee is not None and self.transfer_fee > _MAX_TRANSFER_FEE:
            return f"Must not be greater than {_MAX_TRANSFER_FEE}"
        return None

    def _get_uri_error(self: Self) -> Optional[str]:
        if self.uri is not None and len(self.uri) > _MAX_URI_LENGTH:
            return f"Must not be longer than {_MAX_URI_LENGTH} characters"
        return None



"""Model for NFTokenModify transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Final, Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import HEX_REGEX, KW_ONLY_DATACLASS, require_kwargs_on_init

_MAX_URI_LENGTH: Final[int] = 512


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class NFTokenModify(Transaction):
    """
    The NFTokenModify transaction modifies an NFToken's URI
    if its tfMutable is set to true.
    """

    nftoken_id: str = REQUIRED  # type: ignore
    """
    Identifies the TokenID of the NFToken object that the
    offer references. This field is required.
    """

    owner: Optional[str] = None
    """
    Indicates the AccountID of the account that owns the
    corresponding NFToken.
    """

    uri: Optional[str] = None
    """
    URI that points to the data and/or metadata associated with the NFT.
    This field need not be an HTTP or HTTPS URL; it could be an IPFS URI, a
    magnet link, immediate data encoded as an RFC2379 "data" URL, or even an
    opaque issuer-specific encoding. The URI is not checked for validity.

    This field must be hex-encoded. You can use `xrpl.utils.str_to_hex` to
    convert a UTF-8 string to hex.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.NFTOKEN_MODIFY,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "uri": self._get_uri_error(),
            }.items()
            if value is not None
        }

    def _get_uri_error(self: Self) -> Optional[str]:
        if not self.uri:
            return "URI must not be empty string"
        elif len(self.uri) > _MAX_URI_LENGTH:
            return f"URI must not be longer than {_MAX_URI_LENGTH} characters"

        if not HEX_REGEX.fullmatch(self.uri):
            return "URI must be encoded in hex"
        return None



"""Model for AMMBid transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

from typing_extensions import Final, Self

from xrpl.models.amounts.issued_currency_amount import IssuedCurrencyAmount
from xrpl.models.auth_account import AuthAccount
from xrpl.models.currencies import Currency
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init

_MAX_AUTH_ACCOUNTS: Final[int] = 4


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class AMMBid(Transaction):
    """
    Bid on an Automated Market Maker's (AMM's) auction slot.

    If you win, you can trade against the AMM at a discounted fee until you are outbid
    or 24 hours have passed.
    If you are outbid before 24 hours have passed, you are refunded part of the cost
    of your bid based on how much time remains.
    You bid using the AMM's LP Tokens; the amount of a winning bid is returned
    to the AMM, decreasing the outstanding balance of LP Tokens.
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    The definition for one of the assets in the AMM's pool. This field is required.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    The definition for the other asset in the AMM's pool. This field is required.
    """

    bid_min: Optional[IssuedCurrencyAmount] = None
    """
    Pay at least this LPToken amount for the slot.
    Setting this value higher makes it harder for others to outbid you.
    If omitted, pay the minimum necessary to win the bid.
    """

    bid_max: Optional[IssuedCurrencyAmount] = None
    """
    Pay at most this LPToken amount for the slot.
    If the cost to win the bid is higher than this amount, the transaction fails.
    If omitted, pay as much as necessary to win the bid.
    """

    auth_accounts: Optional[List[AuthAccount]] = None
    """
    A list of up to 4 additional accounts that you allow to trade at the discounted fee.
    This cannot include the address of the transaction sender.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_BID,
        init=False,
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "auth_accounts": self._get_auth_accounts_error(),
            }.items()
            if value is not None
        }

    def _get_auth_accounts_error(self: Self) -> Optional[str]:
        if (
            self.auth_accounts is not None
            and len(self.auth_accounts) > _MAX_AUTH_ACCOUNTS
        ):
            return f"Length must not be greater than {_MAX_AUTH_ACCOUNTS}"
        return None



"""
The tx method retrieves information on a single transaction.

`See tx <https://xrpl.org/tx.html>`_
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class Tx(Request):
    """
    The tx method retrieves information on a single transaction.
    The Request must contain either transaction or CTID parameter, but not both.

    `See tx <https://xrpl.org/tx.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.TX, init=False)
    ctid: Optional[str] = None
    """
    This field is optional.

    The compact transaction identifier of the transaction to look up.
    Must use uppercase hexadecimal only. New in: rippled 1.12.0  (Not supported in
    Clio v2.0 and earlier)

    The tx request accepts either ctid or transaction parameter, but not both.
    """

    transaction: Optional[str] = None
    """
    This field is optional.

    The 256-bit hash of the transaction to look up, as hexadecimal.

    The tx request accepts either ctid or transaction parameter, but not both.
    """

    binary: bool = False
    """
    This field is optional.

    If true, return transaction data and metadata as binary serialized to
    hexadecimal strings. If false, return transaction data and metadata as JSON.
    The default is false.
    """

    min_ledger: Optional[int] = None
    """
    This field is optional.

    Use this with max_ledger to specify a range of up to 1000 ledger indexes, starting
    with this ledger (inclusive). If the server cannot find the transaction, it
    confirms whether it was able to search all the ledgers in this range.
    """

    max_ledger: Optional[int] = None
    """
    This field is optional.

    Use this with min_ledger to specify a range of up to 1000 ledger indexes, ending
    with this ledger (inclusive). If the server cannot find the transaction, it
    confirms whether it was able to search all the ledgers in the requested range.
    """

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if not self._has_only_one_input():
            errors["Tx"] = (
                "Must have only one of `ctid` or `transaction`, but not both."
            )
        return errors

    def _has_only_one_input(self: Self) -> bool:
        unique_ids = [self.transaction, self.ctid]
        present_items = list(filter(bool, unique_ids))
        return len(present_items) == 1



"""Codec for serializing and deserializing issued currency fields."""

from __future__ import annotations

from typing import Any, Dict, Optional, Type, Union

from typing_extensions import Self

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.account_id import AccountID
from xrpl.core.binarycodec.types.currency import Currency
from xrpl.core.binarycodec.types.hash192 import HASH192_BYTES, Hash192
from xrpl.core.binarycodec.types.serialized_type import SerializedType
from xrpl.models.currencies import XRP as XRPModel
from xrpl.models.currencies import IssuedCurrency as IssuedCurrencyModel
from xrpl.models.currencies import MPTCurrency as MPTCurrencyModel


class Issue(SerializedType):
    """Codec for serializing and deserializing issued currency fields."""

    def __init__(self: Self, buffer: bytes) -> None:
        """
        Construct an Issue from given bytes.

        Args:
            buffer: The byte buffer that will be used to store the serialized
                encoding of this field.
        """
        super().__init__(buffer)

    @classmethod
    def from_value(cls: Type[Self], value: Dict[str, str]) -> Self:
        """
        Construct an Issue object from a string or dictionary representation
        of an issued currency.

        Args:
            value: The dictionary to construct an Issue object from.

        Returns:
            An Issue object constructed from value.

        Raises:
            XRPLBinaryCodecException: If the Issue representation is invalid.
        """
        if XRPModel.is_dict_of_model(value):
            currency_bytes = bytes(Currency.from_value(value["currency"]))
            return cls(currency_bytes)

        if IssuedCurrencyModel.is_dict_of_model(value):
            currency_bytes = bytes(Currency.from_value(value["currency"]))
            issuer_bytes = bytes(AccountID.from_value(value["issuer"]))
            return cls(currency_bytes + issuer_bytes)

        if MPTCurrencyModel.is_dict_of_model(value):
            mpt_issuance_id_bytes = bytes(Hash192.from_value(value["mpt_issuance_id"]))
            return cls(bytes(mpt_issuance_id_bytes))

        raise XRPLBinaryCodecException(
            "Invalid type to construct an Issue: expected XRP, IssuedCurrency or "
            f"MPTCurrency as a str or dict, received {value.__class__.__name__}."
        )

    @classmethod
    def from_parser(
        cls: Type[Self],
        parser: BinaryParser,
        length_hint: Optional[int] = None,
    ) -> Self:
        """
        Construct an Issue object from an existing BinaryParser.

        Args:
            parser: The parser to construct the Issue object from.
            length_hint: The number of bytes to consume from the parser.
                For an MPT amount, pass 24 (the fixed length for Hash192).

        Returns:
            The Issue object constructed from a parser.
        """
        # Check if it's an MPTIssue by checking mpt_issuance_id byte size
        if length_hint == HASH192_BYTES:
            mpt_bytes = parser.read(HASH192_BYTES)
            return cls(mpt_bytes)

        currency = Currency.from_parser(parser)
        if currency.to_json() == "XRP":
            return cls(bytes(currency))

        issuer = parser.read(20)  # the length in bytes of an account ID
        return cls(bytes(currency) + issuer)

    def to_json(self: Self) -> Union[str, Dict[Any, Any]]:
        """
        Returns the JSON representation of an issued currency.

        Returns:
            The JSON representation of an Issue.
        """
        # If the buffer is exactly 24 bytes, treat it as an MPT amount.
        if len(self.buffer) == HASH192_BYTES:
            return {"mpt_issuance_id": self.to_hex().upper()}

        parser = BinaryParser(self.to_hex())
        currency: Union[str, Dict[Any, Any]] = Currency.from_parser(parser).to_json()
        if currency == "XRP":
            return {"currency": currency}

        issuer = AccountID.from_parser(parser)
        return {"currency": currency, "issuer": issuer.to_json()}



"""Model for the CredentialDelete transaction"""

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import (
    KW_ONLY_DATACLASS,
    get_credential_type_error,
    require_kwargs_on_init,
)


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class CredentialDelete(Transaction):
    """This transaction deletes a Credential object."""

    account: str = REQUIRED  # type: ignore
    """The transaction submitter."""

    subject: Optional[str] = None
    """The person that the credential is for. If omitted, Account is assumed to be the
    subject."""

    issuer: Optional[str] = None
    """The issuer of the credential. If omitted, Account is assumed to be the issuer."""

    credential_type: str = REQUIRED  # type: ignore
    """A hex-encoded value to identify the type of credential from the issuer."""

    transaction_type: TransactionType = field(
        default=TransactionType.CREDENTIAL_DELETE, init=False
    )

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()

        if (
            cred_type_err := get_credential_type_error(self.credential_type)
        ) is not None:
            errors["credential_type"] = cred_type_err

        if not self.subject and not self.issuer:
            errors["CredentialDelete"] = (
                "either `issuer` or `subject` must be provided."
            )

        return errors



"""This request gets information about an Automated Market Maker (AMM) instance."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Self

from xrpl.models.currencies import Currency
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class AMMInfo(Request):
    """
    The `amm_info` method gets information about an Automated Market Maker
    (AMM) instance.
    """

    amm_account: Optional[str] = None
    """
    The address of the AMM pool to look up.
    """

    asset: Optional[Currency] = None
    """
    One of the assets of the AMM pool to look up.
    """

    asset2: Optional[Currency] = None
    """
    The other asset of the AMM pool.
    """

    method: RequestMethod = field(default=RequestMethod.AMM_INFO, init=False)

    def _get_errors(self: Self) -> Dict[str, str]:
        errors = super()._get_errors()
        if (self.asset is None) != (self.asset2 is None):
            errors["assets"] = "Must have both `asset` and `asset2` fields."
        if (self.asset is None) == (self.amm_account is None):
            errors["params"] = "Must not have both `asset` and `amm_account` fields."
        return errors



"""Class for serializing/deserializing Dicts of objects."""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Type, Union

from typing_extensions import Final, Self

from xrpl.core.addresscodec import is_valid_xaddress, xaddress_to_classic_address
from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.definitions import (
    FieldInstance,
    get_field_instance,
    get_ledger_entry_type_code,
    get_ledger_entry_type_name,
    get_permission_value_type_code,
    get_permission_value_type_name,
    get_transaction_result_code,
    get_transaction_result_name,
    get_transaction_type_code,
    get_transaction_type_name,
)
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.serialized_type import SerializedType
from xrpl.core.binarycodec.types.uint64 import SPECIAL_FIELDS

_OBJECT_END_MARKER_BYTE: Final[bytes] = bytes([0xE1])
_OBJECT_END_MARKER: Final[str] = "ObjectEndMarker"
_ST_OBJECT: Final[str] = "STObject"
_DESTINATION: Final[str] = "Destination"
_ACCOUNT: Final[str] = "Account"
_SOURCE_TAG: Final[str] = "SourceTag"
_DEST_TAG: Final[str] = "DestinationTag"

_UNL_MODIFY_TX: Final[str] = "0066"


def _handle_xaddress(field: str, xaddress: str) -> Dict[str, Union[str, int]]:
    """Break down an X-Address into a classic address and a tag.

    Args:
        field: Name of field
        xaddress: X-Address corresponding to the field

    Returns:
        A dictionary representing the classic address and tag.

    Raises:
        XRPLBinaryCodecException: field-tag combo is invalid.
    """
    (classic_address, tag, is_test_network) = xaddress_to_classic_address(xaddress)
    if field == _DESTINATION:
        tag_name = _DEST_TAG
    elif field == _ACCOUNT:
        tag_name = _SOURCE_TAG
    elif tag is not None:
        raise XRPLBinaryCodecException(f"{field} cannot have an associated tag")

    if tag is not None:
        return {field: classic_address, tag_name: tag}
    return {field: classic_address}


# all of these fields have enum values that are used for serialization
# converts the string name to the corresponding enum code
def _str_to_enum(field: str, value: str) -> Union[str, int]:
    if field == "TransactionType":
        return get_transaction_type_code(value)
    if field == "TransactionResult":
        return get_transaction_result_code(value)
    if field == "LedgerEntryType":
        return get_ledger_entry_type_code(value)
    if field == "PermissionValue":
        return get_permission_value_type_code(value)
    return value


# reverse of the above function
def _enum_to_str(field: str, value: int) -> Union[str, int]:
    if field == "TransactionType":
        return get_transaction_type_name(value)
    if field == "TransactionResult":
        return get_transaction_result_name(value)
    if field == "LedgerEntryType":
        return get_ledger_entry_type_name(value)
    if field == "PermissionValue":
        return get_permission_value_type_name(value)
    return value


class STObject(SerializedType):
    """Class for serializing/deserializing Dicts of objects."""

    @classmethod
    def from_parser(
        cls: Type[Self],
        parser: BinaryParser,
        _length_hint: Optional[None] = None,
    ) -> Self:
        """
        Construct a STObject from a BinaryParser.

        Args:
            parser: The parser to construct a STObject from.

        Returns:
            The STObject constructed from parser.
        """
        from xrpl.core.binarycodec.binary_wrappers.binary_serializer import (
            BinarySerializer,
        )

        serializer = BinarySerializer()

        while not parser.is_end():
            field = parser.read_field()
            if field.name == _OBJECT_END_MARKER:
                break

            associated_value = parser.read_field_value(field)
            serializer.write_field_and_value(field, associated_value)
            if field.type == _ST_OBJECT:
                serializer.append(_OBJECT_END_MARKER_BYTE)

        return cls(bytes(serializer))

    @classmethod
    def from_value(
        cls: Type[Self], value: Dict[str, Any], only_signing: bool = False
    ) -> Self:
        """
        Create a STObject object from a dictionary.

        Args:
            value: The dictionary to construct a STObject from.
            only_signing: whether only the signing fields should be included.

        Returns:
            The STObject object constructed from value.

        Raises:
            XRPLBinaryCodecException: If the STObject can't be constructed
                from value.
        """
        from xrpl.core.binarycodec.binary_wrappers.binary_serializer import (
            BinarySerializer,
        )

        serializer = BinarySerializer()

        xaddress_decoded = {}
        for k, v in value.items():
            if isinstance(v, str) and is_valid_xaddress(v):
                handled = _handle_xaddress(k, v)
                if (
                    _SOURCE_TAG in handled
                    and handled[_SOURCE_TAG] is not None
                    and _SOURCE_TAG in value
                    and value[_SOURCE_TAG] is not None
                    and handled[_SOURCE_TAG] != value[_SOURCE_TAG]
                ):
                    raise XRPLBinaryCodecException(
                        "Cannot have mismatched Account X-Address and SourceTag"
                    )
                if (
                    _DEST_TAG in handled
                    and handled[_DEST_TAG] is not None
                    and _DEST_TAG in value
                    and value[_DEST_TAG] is not None
                    and handled[_DEST_TAG] != value[_DEST_TAG]
                ):
                    raise XRPLBinaryCodecException(
                        "Cannot have mismatched Destination X-Address and "
                        "DestinationTag"
                    )
                xaddress_decoded.update(handled)
            else:
                xaddress_decoded[k] = _str_to_enum(k, v)

        sorted_keys: List[FieldInstance] = []
        for field_name in xaddress_decoded:
            field_instance = get_field_instance(field_name)
            if (
                field_instance is not None
                and xaddress_decoded[field_instance.name] is not None
                and field_instance.is_serialized
            ):
                sorted_keys.append(field_instance)
        sorted_keys.sort(key=lambda x: x.ordinal)

        if only_signing:
            sorted_keys = list(filter(lambda x: x.is_signing, sorted_keys))

        is_unl_modify = False

        for field in sorted_keys:
            try:
                args = (
                    (xaddress_decoded[field.name], field.name)
                    if field.name in SPECIAL_FIELDS
                    else (xaddress_decoded[field.name],)
                )
                associated_value = field.associated_type.from_value(*args)
            except XRPLBinaryCodecException as e:
                # mildly hacky way to get more context in the error
                # provides the field name and not just the type it's expecting
                # keeps the original stack trace
                e.args = (f"Error processing {field.name}: {e.args[0]}",) + e.args[1:]
                raise
            if (
                field.name == "TransactionType"
                and str(associated_value) == _UNL_MODIFY_TX
            ):
                # triggered when the TransactionType field has a value of 'UNLModify'
                is_unl_modify = True
            is_unl_modify_workaround = field.name == "Account" and is_unl_modify
            # true when in the UNLModify pseudotransaction (after the transaction type
            # has been processed) and working with the Account field
            # The Account field must not be a part of the UNLModify pseudotransaction
            # encoding, due to a bug in rippled

            serializer.write_field_and_value(
                field, associated_value, is_unl_modify_workaround
            )
            if field.type == _ST_OBJECT:
                serializer.append(_OBJECT_END_MARKER_BYTE)

        return cls(bytes(serializer))

    def to_json(self: Self) -> Dict[str, Any]:
        """
        Returns the JSON representation of a STObject.

        Returns:
            The JSON representation of a STObject.
        """
        parser = BinaryParser(str(self))
        accumulator = {}

        while not parser.is_end():
            field = parser.read_field()
            if field.name == _OBJECT_END_MARKER:
                break
            json_value = parser.read_field_value(field).to_json()
            accumulator[field.name] = _enum_to_str(field.name, json_value)

        return accumulator



"""Class for serializing and deserializing Lists of objects.
See `Array Fields <https://xrpl.org/serialization.html#array-fields>`_
"""

from __future__ import annotations

from typing import Any, List, Optional, Type

from typing_extensions import Final, Self

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.serialized_type import SerializedType
from xrpl.core.binarycodec.types.st_object import STObject

_ARRAY_END_MARKER: Final[bytes] = bytes([0xF1])
_ARRAY_END_MARKER_NAME: Final[str] = "ArrayEndMarker"

_OBJECT_END_MARKER: Final[bytes] = bytes([0xE1])


class STArray(SerializedType):
    """Class for serializing and deserializing Lists of objects.
    See `Array Fields <https://xrpl.org/serialization.html#array-fields>`_
    """

    @classmethod
    def from_parser(
        cls: Type[Self],
        parser: BinaryParser,
        _length_hint: Optional[None] = None,
    ) -> Self:
        """
        Construct a STArray from a BinaryParser.

        Args:
            parser: The parser to construct a STArray from.

        Returns:
            The STArray constructed from parser.
        """
        bytestring = b""

        while not parser.is_end():
            field = parser.read_field()
            if field.name == _ARRAY_END_MARKER_NAME:
                break
            bytestring += bytes(field.header)
            bytestring += bytes(parser.read_field_value(field))
            bytestring += _OBJECT_END_MARKER

        bytestring += _ARRAY_END_MARKER
        return cls(bytestring)

    @classmethod
    def from_value(cls: Type[Self], value: List[Any]) -> Self:
        """
        Create a STArray object from a dictionary.

        Args:
            value: The dictionary to construct a STArray from.

        Returns:
            The STArray object constructed from value.

        Raises:
            XRPLBinaryCodecException: If the provided value isn't a list or contains
                non-dict elements.
        """
        if not isinstance(value, list):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a STArray:"
                " expected list, received {value.__class__.__name__}."
            )

        if len(value) > 0 and not isinstance(value[0], dict):
            raise XRPLBinaryCodecException(
                ("Cannot construct STArray from a list of non-dict" " objects")
            )

        bytestring = b""
        for obj in value:
            transaction = STObject.from_value(obj)
            bytestring += bytes(transaction)
        bytestring += _ARRAY_END_MARKER
        return cls(bytestring)

    def to_json(self: Self) -> List[Any]:
        """
        Returns the JSON representation of a STArray.

        Returns:
            The JSON representation of a STArray.
        """
        result = []
        parser = BinaryParser(str(self))

        while not parser.is_end():
            field = parser.read_field()
            if field.name == _ARRAY_END_MARKER_NAME:
                break

            outer = {}
            outer[field.name] = STObject.from_parser(parser).to_json()
            result.append(outer)
        return result



"""
The deposit_authorized command indicates whether one account
is authorized to send payments directly to another. See
Deposit Authorization for information on how to require
authorization to deliver money to your account.
"""

from dataclasses import dataclass, field
from typing import List, Optional

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class DepositAuthorized(Request, LookupByLedgerRequest):
    """
    The deposit_authorized command indicates whether one account
    is authorized to send payments directly to another. See
    Deposit Authorization for information on how to require
    authorization to deliver money to your account.
    """

    source_account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    destination_account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    method: RequestMethod = field(default=RequestMethod.DEPOSIT_AUTHORIZED, init=False)

    credentials: Optional[List[str]] = None
    """List of Credential ID strings. If this field is included, then the
    credential will be taken into account when analyzing whether the sender can send
    funds to the destination.
    """
